//================================================================================================
//                                  cIotModuleCom body
//================================================================================================

void cIotModuleCom::Initialize(void *pContext, pWiFiEventHandler WiFiEventHandler)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  //  save context
  pIotModuleContext = pContext;
  WiFiEventHandlerCallback = WiFiEventHandler;
  
  WIFI_MODE WiFiMode = WIFI_MODE::STATION;
  if (strcasecmp(IOTMODULE_WIFI_MODE, pSOFTAP) == 0) WiFiMode = WIFI_MODE::SOFTAP;

  //  Chip Silicon package
  switch ((REG_READ(EFUSE_BLK0_RDATA3_REG) >>  EFUSE_RD_CHIP_VER_PKG_S) & EFUSE_RD_CHIP_VER_PKG_V)
  {
    case EFUSE_RD_CHIP_VER_PKG_ESP32D0WDQ6:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D0WDQ6", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32D0WDQ5:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D0WDQ5", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32D2WDQ5:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : D2WDQ5", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32PICOD2:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : PICOD2", __FILENAME__, __LINE__);
      break;
    case EFUSE_RD_CHIP_VER_PKG_ESP32PICOD4:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip package : PICOD4", __FILENAME__, __LINE__);
      break;
  }

  //  Chip Silicon Revision
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Chip Silicon Revision : %d", __FILENAME__, __LINE__, (REG_READ(EFUSE_BLK0_RDATA3_REG) >>  EFUSE_RD_CHIP_VER_REV1_S) & EFUSE_RD_CHIP_VER_REV1_V);

  // typedef struct {
  //    rtc_cpu_freq_src_t source;      //!< The clock from which CPU clock is derived
  //    uint32_t source_freq_mhz;       //!< Source clock frequency
  //    uint32_t div;                   //!< Divider, freq_mhz = source_freq_mhz / div
  //    uint32_t freq_mhz;              //!< CPU clock frequency
  // } rtc_cpu_freq_config_t;
  rtc_cpu_freq_config_t rtc_cpu_freq_config;
  rtc_clk_cpu_freq_get_config(&rtc_cpu_freq_config);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - CPU cores frequency : %d Mhz", __FILENAME__, __LINE__, rtc_cpu_freq_config.freq_mhz);
  for (uint8_t i = 0; i < 2; i++)
  {
    /*
    POWERON_RESET           1   Vbat power on reset
    SW_RESET                3   Software reset digital core
    OWDT_RESET              4   Legacy watch dog reset digital core
    DEEPSLEEP_RESET         3   Deep Sleep reset digital core
    SDIO_RESET              6   Reset by SLC module, reset digital core
    TG0WDT_SYS_RESET        7   Timer Group0 Watch dog reset digital core
    TG1WDT_SYS_RESET        8   Timer Group1 Watch dog reset digital core
    RTCWDT_SYS_RESET        9   RTC Watch dog Reset digital core
    INTRUSION_RESET         10  Instrusion tested to reset CPU
    TGWDT_CPU_RESET         11  Time Group reset CPU
    SW_CPU_RESET            12  Software reset CPU
    RTCWDT_CPU_RESET        13  RTC Watch dog Reset CPU
    EXT_CPU_RESET           14  for APP CPU, reseted by PRO CPU
    RTCWDT_BROWN_OUT_RESET  15  Reset when the vdd voltage is not stable
    RTCWDT_RTC_RESET        16  RTC Watch dog reset digital core and rtc module 

    -PRO-  -APP- ------Source------- ----ResetType---

    0x01   0x01  Chip Power On Reset System Reset
    0x10   0x10  RWDT System Reset   System Reset
    0x0F   0x0F  Brown Out Reset     System Reset
    0x03   0x03  Software System     Reset Core Reset
    0x05   0x05  Deep Sleep Reset    Core Reset
    0x07   0x07  MWDT0 Global Reset  Core Reset
    0x08   0x08  MWDT1 Global Reset  Core Reset
    0x09   0x09  RWDT Core Reset     Core Reset
    0x0B         MWDT0 CPU Reset     CPU Reset
    0x0C         Software CPU Reset  CPU Reset
           0x0B  MWDT1 CPU Reset     CPU Reset
           0x0C  Software CPU Reset  CPU Reset
    0x0D   0x0D  RWDT CPU Reset      CPU Reset 
           0x0E  PRO CPU Reset       CPU Reset
    */

    const char *pReason[] =
    {
      "NO_MEAN",
      "POWERON_RESET",
      "NO_MEAN",
      "SW_RESET",
      "OWDT_RESET",
      "DEEPSLEEP_RESET",
      "SDIO_RESET",
      "TG0WDT_SYS_RESET",
      "TG1WDT_SYS_RESET",
      "RTCWDT_SYS_RESET",
      "INTRUSION_RESET",
      "TGWDT_CPU_RESET",
      "SW_CPU_RESET",
      "RTCWDT_CPU_RESET",
      "EXT_CPU_RESET",
      "RTCWDT_BROWN_OUT_RESET",
      "RTCWDT_RTC_RESET",
    };
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Reset reason core %d : %s", __FILENAME__, __LINE__, i, pReason[rtc_get_reset_reason(i)]);
  }

  //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Wakeup reason : %s", __FILENAME__, __LINE__, pReason[rtc_get_wakeup_cause()]);

  //  Cache enabled
  if (spi_flash_cache_enabled()) ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - both CPUs have flash cache enabled", __FILENAME__, __LINE__);

  size_t ul;
  ul = spi_flash_get_chip_size();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Flash chip size : %d bytes", __FILENAME__, __LINE__, ul);

  //=====================================================================================
  //  flash mapped regions
  //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Flash mapped regions", __FILENAME__, __LINE__);
  //  this function use always serial monitor to dump info, is not ESP_LOG frendly
  //  spi_flash_mmap_dump();
  //=====================================================================================
  //  app_main() -> initArduino() already re-initialize NVS in case of ESP_ERR_NVS_NO_FREE_PAGES
  //  calling esp_partition_find_first(NVS) and esp_partition_erase_range()
  //=====================================================================================
  //  initialize NVS.
  //  esp_err_t err = nvs_flash_init();
  //  if (err == ESP_ERR_NVS_NO_FREE_PAGES /*|| err == ESP_ERR_NVS_NEW_VERSION_FOUND*/)
  //  {
  //    //  OTA app partition table has a smaller NVS partition size than the non-OTA
  //    //  partition table. This size mismatch may cause NVS initialization to fail.
  //    //  If this happens, we erase NVS partition and initialize NVS again.
  //    //  We erase it also if new NVS version is available
  //    ESP_ERROR_CHECK(nvs_flash_erase());
  //    err = nvs_flash_init();
  //  }
  //  //  after init or erase & init there must be NO error left
  //  ESP_ERROR_CHECK(err);
  //=====================================================================================

  //  partitions
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Partition table:", __FILENAME__, __LINE__);
  esp_partition_iterator_t _mypartiterator;
  const esp_partition_t *_mypart;
  _mypartiterator = esp_partition_find(ESP_PARTITION_TYPE_APP, ESP_PARTITION_SUBTYPE_ANY, NULL);
  if (_mypartiterator)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - App Partition table:", __FILENAME__, __LINE__);
    do
    {
      _mypart = esp_partition_get(_mypartiterator);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Type: %02x SubType %02x Address 0x%06X Size 0x%06X Encryption %i Label %s", __FILENAME__, __LINE__, _mypart->type, _mypart->subtype, _mypart->address, _mypart->size, _mypart->encrypted, _mypart->label);
    }
    while ((_mypartiterator = (esp_partition_next(_mypartiterator))));
  }
  esp_partition_iterator_release(_mypartiterator);
  _mypartiterator = esp_partition_find(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_ANY, NULL);
  if (_mypartiterator)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Data Partition table:", __FILENAME__, __LINE__);
    do
    {
      _mypart = esp_partition_get(_mypartiterator);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Type: %02x SubType %02x Address 0x%06X Size 0x%06X Encryption %i Label %s", __FILENAME__, __LINE__, _mypart->type, _mypart->subtype, _mypart->address, _mypart->size, _mypart->encrypted, _mypart->label);
    }
    while ((_mypartiterator = esp_partition_next(_mypartiterator))) ;
  }
  esp_partition_iterator_release(_mypartiterator);

  //  Running partition
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Running partition : %s", __FILENAME__, __LINE__, esp_ota_get_running_partition()->label);

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Arduino IDE ESP-IDF : %s", __FILENAME__, __LINE__, esp_get_idf_version());
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ESP-IDF portTICK_PERIOD_MS : %d", __FILENAME__, __LINE__, portTICK_PERIOD_MS);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_FIRMWARE_VER : %d", __FILENAME__, __LINE__, IOTMODULE_FIRMWARE_VER);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_THEREMINO_IOT_PROTOCOL : %s", __FILENAME__, __LINE__, IOTMODULE_THEREMINO_IOT_PROTOCOL);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - internal IOTMODULE_THEREMINO_IOT_PROTOCOL_VER : %d", __FILENAME__, __LINE__, IOTMODULE_THEREMINO_IOT_PROTOCOL_VER);
  if (IOTMODULE_OUTPUT_WATCHDOG_ENABLE == 1)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_OUTPUT_WATCHDOG_ENABLE : %d", __FILENAME__, __LINE__, IOTMODULE_OUTPUT_WATCHDOG_ENABLE);
  }
  if (IOTMODULE_FUOTA_ENABLE == 1)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_FUOTA_ENABLE : %d", __FILENAME__, __LINE__, IOTMODULE_FUOTA_ENABLE);
  }

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - configured IOTMODULE_WIFI_MODE : %s", __FILENAME__, __LINE__, WiFiMode == WIFI_MODE::STATION ? "Station" : "SoftAP");
  
  //  if configured WiFi mode must be overriden by a pin status
  if (IotModuleApp.WiFiModeOverridePinDefined())
  {
    //  override WiFi mode
    WiFiMode = IotModuleApp.WiFiModeOverridePinStatus() == false ? WIFI_MODE::STATION : WIFI_MODE::SOFTAP;
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IOTMODULE_WIFI_MODE_PIN overrides : IOTMODULE_WIFI_MODE : %s", __FILENAME__, __LINE__, WiFiMode == WIFI_MODE::STATION ? "Station" : "SoftAP");
  }

  //  event group is used by ComServerTask and WiFiEventHandler
  hEventGroup = xEventGroupCreate();
  //  get current execution core ID
  int32_t CoreID = xPortGetCoreID();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Arduino setup() and loop() execution core : %d", __FILENAME__, __LINE__, CoreID);
  //  if nedeed create ComServerTask pinned to different core
  CoreID = (CoreID == 0 ? 0 : 1);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ComServerTask execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());
  if (xTaskCreatePinnedToCore(ComServerTask, "udp_server", COM_TASK_STACK_SIZE, this, tskIDLE_PRIORITY + 6, &hComServerTask, CoreID) == pdPASS)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - ComServerTask created", __FILENAME__, __LINE__);
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ComServerTask not created", __FILENAME__, __LINE__);
  }

  //  IotModuleApp initialization must be done before any wifi loop event
  IotModuleApp.Initialize();

  tcpip_adapter_init();

  const char pWiFiAssoc[] = IOTMODULE_WIFI_ASSOC;
  const char pWiFiSSID[] = IOTMODULE_WIFI_SSID;
  const char pWiFiPassword[] = IOTMODULE_WIFI_PASSWORD;
  const char pIP_Mode[] = IOTMODULE_IP_MODE;
  const char pIP_IP[] = IOTMODULE_IP_IP;
  const char pIP_Network[] = IOTMODULE_IP_NETWORK;
  const char pIP_Gateway[] = IOTMODULE_IP_GATEWAY;
  const char pIP_DNS[] = IOTMODULE_IP_DNS;

  //  split WiFiAssoc list
  std::stringstream StreamWiFiAssoc(pWiFiAssoc);
  std::vector<std::string> TokensWiFiAssoc;
  while (StreamWiFiAssoc)
  {
    std::string word;
    StreamWiFiAssoc >> word;
    if (word.length() != 0) TokensWiFiAssoc.push_back(word);
  }
  uint8_t WiFiAssoc_Size = TokensWiFiAssoc.size();

  //  split WiFiSSID list
  std::stringstream StreamWiFiSSID(pWiFiSSID);
  std::vector<std::string> TokensWiFiSSID;
  while (StreamWiFiSSID)
  {
    std::string word;
    StreamWiFiSSID >> word;
#ifdef IOTMODULE_SPACE_CHAR    
    replace(word.begin(), word.end(), IOTMODULE_SPACE_CHAR, ' ');
#endif    
    if (word.length() != 0) TokensWiFiSSID.push_back(word);
  }
  uint8_t WiFiSSID_Size = TokensWiFiSSID.size();

  //  split WiFiPassword list
  std::stringstream StreamWiFiPassword(pWiFiPassword);
  std::vector<std::string> TokensWiFiPassword;
  while (StreamWiFiPassword)
  {
    std::string word;
    StreamWiFiPassword >> word;
#ifdef IOTMODULE_SPACE_CHAR    
    replace(word.begin(), word.end(), IOTMODULE_SPACE_CHAR, ' ');
#endif    
    if (word.length() != 0) TokensWiFiPassword.push_back(word);
  }
  uint8_t WiFiPassword_Size = TokensWiFiPassword.size();

  //  split IP_Mode list
  std::stringstream StreamIP_Mode(pIP_Mode);
  std::vector<std::string> TokensIP_Mode;
  while (StreamIP_Mode)
  {
    std::string word;
    StreamIP_Mode >> word;
    if (word.length() != 0) TokensIP_Mode.push_back(word);
  }
  uint8_t IP_Mode_Size = TokensIP_Mode.size();

  //  split IP_IP list
  std::stringstream StreamIP_IP(pIP_IP);
  std::vector<std::string> TokensIP_IP;
  while (StreamIP_IP)
  {
    std::string word;
    StreamIP_IP >> word;
    if (word.length() != 0) TokensIP_IP.push_back(word);
  }
  uint8_t IP_IP_Size = TokensIP_IP.size();

  //  split IP_Network list
  std::stringstream StreamIP_Network(pIP_Network);
  std::vector<std::string> TokensIP_Network;
  while (StreamIP_Network)
  {
    std::string word;
    StreamIP_Network >> word;
    if (word.length() != 0) TokensIP_Network.push_back(word);
  }
  uint8_t IP_Network_Size = TokensIP_Network.size();

  //  split IP_Gateway list
  std::stringstream StreamIP_Gateway(pIP_Gateway);
  std::vector<std::string> TokensIP_Gateway;
  while (StreamIP_Gateway)
  {
    std::string word;
    StreamIP_Gateway >> word;
    if (word.length() != 0) TokensIP_Gateway.push_back(word);
  }
  uint8_t IP_Gateway_Size = TokensIP_Gateway.size();

  //  split IP_DNS list
  std::stringstream StreamIP_DNS(pIP_DNS);
  std::vector<std::string> TokensIP_DNS;
  while (StreamIP_DNS)
  {
    std::string word;
    StreamIP_DNS >> word;
    if (word.length() != 0) TokensIP_DNS.push_back(word);
  }
  uint8_t IP_DNS_Size = TokensIP_DNS.size();

  //  count colums
  uint8_t CommonColumnSizeMin = WiFiAssoc_Size;
  if (WiFiSSID_Size < CommonColumnSizeMin) CommonColumnSizeMin = WiFiSSID_Size;
  if (WiFiPassword_Size < CommonColumnSizeMin) CommonColumnSizeMin = WiFiPassword_Size;
  if (IP_Mode_Size < CommonColumnSizeMin) CommonColumnSizeMin = IP_Mode_Size;
  if (IP_IP_Size < CommonColumnSizeMin) CommonColumnSizeMin = IP_IP_Size;
  if (IP_Network_Size < CommonColumnSizeMin) CommonColumnSizeMin = IP_Network_Size;
  if (IP_Gateway_Size < CommonColumnSizeMin) CommonColumnSizeMin = IP_Gateway_Size;
  if (IP_DNS_Size < CommonColumnSizeMin) CommonColumnSizeMin = IP_DNS_Size;
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - CommonColumnSizeMin : %d", __FILENAME__, __LINE__, CommonColumnSizeMin);
  uint8_t CommonColumnSizeMax = WiFiAssoc_Size;
  if (WiFiSSID_Size > CommonColumnSizeMax) CommonColumnSizeMax = WiFiSSID_Size;
  if (WiFiPassword_Size > CommonColumnSizeMax) CommonColumnSizeMax = WiFiPassword_Size;
  if (IP_Mode_Size > CommonColumnSizeMax) CommonColumnSizeMax = IP_Mode_Size;
  if (IP_IP_Size > CommonColumnSizeMax) CommonColumnSizeMax = IP_IP_Size;
  if (IP_Network_Size > CommonColumnSizeMax) CommonColumnSizeMax = IP_Network_Size;
  if (IP_Gateway_Size > CommonColumnSizeMax) CommonColumnSizeMax = IP_Gateway_Size;
  if (IP_DNS_Size > CommonColumnSizeMax) CommonColumnSizeMax = IP_DNS_Size;
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - CommonColumnSizeMax : %d", __FILENAME__, __LINE__, CommonColumnSizeMax);
  uint8_t CommonColumnSize = CommonColumnSizeMin;
  if (CommonColumnSizeMin != CommonColumnSizeMax)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - CommonColumnSizeMin != CommonColumnSizeMax", __FILENAME__, __LINE__);
  }

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - total number of valid configuration columns : %d", __FILENAME__, __LINE__, CommonColumnSize);

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------");
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "|CONF|         configured SSID          | WiFi assoc | IP mode |");
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------");
  for (int i = 0; i < CommonColumnSize; i++)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "| %2d | %32s |  %8s  | %7s |", i + 1, TokensWiFiSSID[i].c_str(), TokensWiFiAssoc[i].c_str(), TokensIP_Mode[i].c_str());
  }
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------");

  Displayed1 = false;
  Displayed2 = false;
  
  ESP_ERROR_CHECK(esp_event_loop_init(cIotModuleCom::WiFiEventHandler, this));
  ESP_ERROR_CHECK(esp_wifi_init(&wifi_init_config));
  ESP_ERROR_CHECK(esp_wifi_set_storage(WIFI_STORAGE_RAM));
  ESP_ERROR_CHECK(esp_wifi_set_ps(WIFI_PS_NONE));

  //  MAC address
  if (WiFiMode == WIFI_MODE::STATION)
  {
    ESP_ERROR_CHECK(esp_wifi_get_mac(WIFI_IF_STA, WiFiMAC));
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Station mode MAC address : %02X%02X%02X%02X%02X%02X", __FILENAME__, __LINE__, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  }
  else if (WiFiMode == WIFI_MODE::SOFTAP)
  {
    ESP_ERROR_CHECK(esp_wifi_get_mac(WIFI_IF_AP, WiFiMAC));
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - AP mode MAC address : %02X%02X%02X%02X%02X%02X", __FILENAME__, __LINE__, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  }

  if (WiFiMode == WIFI_MODE::STATION)
  {
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_STA ));
  }
  else if (WiFiMode == WIFI_MODE::SOFTAP)
  {
    //  search keyword
    uint8_t i;
    bool Matched = false;
    for (i = 0; i < CommonColumnSize && !Matched; i++)
    {
      if (strcasecmp(TokensWiFiAssoc[i].c_str(), pSOFTAP) == 0)
      {
        Matched = true;
        break;
      } 
    }
    //  if keyword found
    if (Matched)
    {
      //  set configured SSID/password
      strcpy(reinterpret_cast<char*>(wifi_config.ap.ssid), TokensWiFiSSID[i].c_str());
      strcpy((char*)wifi_config.ap.password, TokensWiFiPassword[i].c_str());
    }
    else
    {
      //  set a fallback SSID/password
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - SoftAP configuration (SSID/password) is missing, using (NoName/!!NoName!!)", __FILENAME__, __LINE__);
      strcpy(reinterpret_cast<char*>(wifi_config.ap.ssid), "NoName");
      strcpy((char*)wifi_config.ap.password, "!!NoName!!");
    }

    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - SoftAP active with SSID set to %s", __FILENAME__, __LINE__, wifi_config.ap.ssid);
    
    //  prepare AP configuration data
    wifi_config.ap.ssid_len = 0;
    wifi_config.ap.authmode = WIFI_AUTH_WPA2_PSK;
    wifi_config.ap.max_connection = 1;
    wifi_config.ap.channel = IOTMODULE_WIFI_CHANNEL;
    wifi_config.ap.ssid_hidden = 0;
    wifi_config.ap.beacon_interval = 100;
    ESP_ERROR_CHECK(esp_wifi_set_mode(WIFI_MODE_AP ));
    ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_AP, &wifi_config));
  }
  
  //  start wifi core
  ESP_ERROR_CHECK(esp_wifi_start());

  int8_t TX_PowerFactor;
  uint8_t TX_PowerIndex;

  //  get iotmodule TX power
  ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&TX_PowerFactor));
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IotModule default TX Power : %2.1f dBm", __FILENAME__, __LINE__, TX_PowerFactor * 0.25);

  //  set iotmodule TX power
  TX_PowerIndex = IOTMODULE_WIFI_TX_POWER_INDEX;
  ESP_ERROR_CHECK(esp_wifi_set_max_tx_power(TX_PowerFactorArray[TX_PowerIndex]));
    
  //  verify iotmodule TX power
  ESP_ERROR_CHECK(esp_wifi_get_max_tx_power(&TX_PowerFactor));
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IotModule configured TX Power : %2.1f dBm", __FILENAME__, __LINE__, TX_PowerFactor * 0.25);

  if (WiFiMode == WIFI_MODE::STATION)
  {
    //  loop trough the SSID lists until one of the configured and scanned pairs matches
    bool Matched = false;
    uint16_t ap_num = sizeof(wifi_ap_record) / sizeof(wifi_ap_record_t);
    uint8_t MatchIndex;
    //  loop until matched
    while (!Matched)
    {
      //  if more then one column configured
      if (CommonColumnSize > 1)
      {
        //  scan the neighborhood SSIDs
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Starting WiFi scan", __FILENAME__, __LINE__);
        wifi_scan_config.ssid = 0;
        wifi_scan_config.bssid = 0;
        wifi_scan_config.channel = 0;
        wifi_scan_config.show_hidden = true;
        wifi_scan_config.scan_type = WIFI_SCAN_TYPE_PASSIVE;
        //  todo incremental passive scan time?
        wifi_scan_config.scan_time.passive = 150;
        ESP_ERROR_CHECK(esp_wifi_scan_start(&wifi_scan_config, true));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Scan completed", __FILENAME__, __LINE__);
        //  get updated SSID list
        ESP_ERROR_CHECK(esp_wifi_scan_get_ap_records(&ap_num, wifi_ap_record));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Total number of SSID recognized : %d", __FILENAME__, __LINE__, ap_num);
        //  print scan list
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------------");
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "|SCAN|               SSID               | Channel | RSSI | Auth Mode |");
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------------");
        for (int i = 0; i < ap_num; i++)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "| %2d | %32s |   %3d   | %4d |     %d     |", i + 1, (char *)wifi_ap_record[i].ssid, wifi_ap_record[i].primary, wifi_ap_record[i].rssi, wifi_ap_record[i].authmode);
        }
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "----------------------------------------------------------------------");
        //  for every configured SSID
        for (uint8_t i = 0; i < CommonColumnSize && !Matched; i++)
        {
          //  for every found SSID
          for (uint8_t j = 0; j < ap_num; j++)
          {
            //  if they match and Assoc is STATION
            if (strcmp(TokensWiFiSSID[i].c_str(), (const char*)wifi_ap_record[j].ssid) == 0 && strcasecmp(TokensWiFiAssoc[i].c_str(), pSTATION) == 0)
            {
              //  copy SSID of target wifi network
              strcpy(reinterpret_cast<char*>(wifi_config.sta.ssid), TokensWiFiSSID[i].c_str());
              
              //  copy password
              strcpy((char*)wifi_config.sta.password, TokensWiFiPassword[i].c_str());
              
              //  Do fast scan, scan will end after find SSID match AP 
              wifi_config.sta.scan_method = WIFI_FAST_SCAN;
              //  All channel scan, scan will end after scan all the channel
              //wifi_config.sta.scan_method = WIFI_ALL_CHANNEL_SCAN;

              //  use bssid
              wifi_config.sta.bssid_set = 0;

              //  MAC address of target AP
              //  not used because wifi_config.sta.bssid_set = 0            
              //  uint8_t bssid[6];
              
              //  set channel to AP primary channel
              wifi_config.sta.channel = wifi_ap_record[j].primary;
              
              //  Listen interval for ESP32 station to receive beacon when WIFI_PS_MAX_MODEM is set. Units: AP beacon intervals. Defaults to 3 if set to 0.
              //  This is not our case, we use WIFI_PS_NONE
              //  ??? uint16_t listen_interval;
              
              //  sort the connect AP in the list by rssi or security mode
              wifi_config.sta.sort_method = WIFI_CONNECT_AP_BY_SIGNAL;
              
              //  When scan_method is set, only APs which have an auth mode that is more secure than the selected auth mode and a signal stronger than the minimum RSSI will be used.
              //  When scan_method is set it means when WIFI_ALL_CHANNEL_SCAN set?????
              //  ??? wifi_scan_threshold_t  threshold;
              
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - SSID match between CONF[%d] and SCAN[%d]", __FILENAME__, __LINE__, i + 1, j + 1);
              Matched = true;
              MatchIndex = i;
              break;
            }
          }
        }
      }
      //  if only one column configured speed up avoiding scan the neighborhood SSIDs
      else
      {
        //  copy SSID of target wifi network
        strcpy(reinterpret_cast<char*>(wifi_config.sta.ssid), TokensWiFiSSID[0].c_str());
        
        //  copy password
        strcpy((char*)wifi_config.sta.password, TokensWiFiPassword[0].c_str());
        
        //  Do fast scan, scan will end after find SSID match AP 
        wifi_config.sta.scan_method = WIFI_FAST_SCAN;
        //  All channel scan, scan will end after scan all the channel
        //wifi_config.sta.scan_method = WIFI_ALL_CHANNEL_SCAN;

        //  use bssid
        wifi_config.sta.bssid_set = 0;

        //  MAC address of target AP
        //  not used because wifi_config.sta.bssid_set = 0            
        //  uint8_t bssid[6];
        
        //  set channel to AP primary channel
        wifi_config.sta.channel = 0;
        
        //  Listen interval for ESP32 station to receive beacon when WIFI_PS_MAX_MODEM is set. Units: AP beacon intervals. Defaults to 3 if set to 0.
        //  This is not our case, we use WIFI_PS_NONE
        //  ??? uint16_t listen_interval;
        
        //  sort the connect AP in the list by rssi or security mode
        wifi_config.sta.sort_method = WIFI_CONNECT_AP_BY_SIGNAL;
        
        //  When scan_method is set, only APs which have an auth mode that is more secure than the selected auth mode and a signal stronger than the minimum RSSI will be used.
        //  When scan_method is set it means when WIFI_ALL_CHANNEL_SCAN set?????
        //  ??? wifi_scan_threshold_t  threshold;
        
        Matched = true;
        MatchIndex = 0;
        break;
      }
      if (!Matched) ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - SSID match not found, restarting scan...", __FILENAME__, __LINE__);
    }
    
    IP_MODE IP_Mode = IP_MODE::DHCP;
    if (strcasecmp(TokensIP_Mode[MatchIndex].c_str(), pDHCP) == 0) IP_Mode = IP_MODE::DHCP;
    if (strcasecmp(TokensIP_Mode[MatchIndex].c_str(), pSTATIC) == 0) IP_Mode = IP_MODE::STATIC;

    if (IP_Mode == IP_MODE::STATIC)
    {
      //  check IP
      if (ip4addr_aton(TokensIP_IP[MatchIndex].c_str(), &tcpip_adapter_ip_info.ip) == 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Static IP - configured IP invalid / not specified, set to 1.1.1.1", __FILENAME__, __LINE__);
        ip4addr_aton("1.1.1.1", &tcpip_adapter_ip_info.ip);
      }
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - IP : %s", __FILENAME__, __LINE__, ip4addr_ntoa(&tcpip_adapter_ip_info.ip));
      //  check Network
      if (ip4addr_aton(TokensIP_Network[MatchIndex].c_str(), &tcpip_adapter_ip_info.netmask) == 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Static IP - configured Network invalid / not specified, set to 255.255.255.255", __FILENAME__, __LINE__);
        ip4addr_aton("1.1.1.1", &tcpip_adapter_ip_info.netmask);
      }
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - Network : %s", __FILENAME__, __LINE__, ip4addr_ntoa(&tcpip_adapter_ip_info.netmask));
      //  check Gateway
      if (ip4addr_aton(TokensIP_Gateway[MatchIndex].c_str(), &tcpip_adapter_ip_info.gw) == 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - configured Gateway invalid / not specified, set to 1.1.1.1", __FILENAME__, __LINE__);
        ip4addr_aton("1.1.1.1", &tcpip_adapter_ip_info.gw);
      }
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - Gateway : %s", __FILENAME__, __LINE__, ip4addr_ntoa(&tcpip_adapter_ip_info.gw));
      //  check DNS
      if (ipaddr_aton(TokensIP_DNS[MatchIndex].c_str(), &tcpip_adapter_dns_info.ip) == 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - configured DNS invalid / not specified, set to 1.1.1.1", __FILENAME__, __LINE__);
        ipaddr_aton("1.1.1.1", &tcpip_adapter_dns_info.ip);
      }
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Static IP - DNS : %s", __FILENAME__, __LINE__, ipaddr_ntoa(&tcpip_adapter_dns_info.ip));

      if (WiFiMode == WIFI_MODE::STATION)
      {
        //  stop DHCP client
        ESP_ERROR_CHECK(tcpip_adapter_dhcpc_stop(TCPIP_ADAPTER_IF_STA));
        //  set static IP
        ESP_ERROR_CHECK(tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_STA, &tcpip_adapter_ip_info));
        //  set DNS IP
        ESP_ERROR_CHECK(tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &tcpip_adapter_dns_info));
      }
      else if (WiFiMode == WIFI_MODE::SOFTAP)
      {
        //  stop DHCP server
        ESP_ERROR_CHECK(tcpip_adapter_dhcps_stop(TCPIP_ADAPTER_IF_AP));
        //  set static IP
        ESP_ERROR_CHECK(tcpip_adapter_set_ip_info(TCPIP_ADAPTER_IF_AP, &tcpip_adapter_ip_info));
        //  set DNS IP
        ESP_ERROR_CHECK(tcpip_adapter_set_dns_info(TCPIP_ADAPTER_IF_AP, TCPIP_ADAPTER_DNS_MAIN, &tcpip_adapter_dns_info));
        //  start DHCP server
        ESP_ERROR_CHECK(tcpip_adapter_dhcps_start(TCPIP_ADAPTER_IF_AP));
      }
    }
    else if (IP_Mode == IP_MODE::DHCP)
    {
      //  nothing to do
    }
    ESP_ERROR_CHECK(esp_wifi_set_config(WIFI_IF_STA, &wifi_config));
    ESP_ERROR_CHECK(esp_wifi_connect());
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Connection to access point with SSID %s and %s configuration has been initiated", __FILENAME__, __LINE__, wifi_config.sta.ssid, IP_Mode == IP_MODE::DHCP ? pDHCP : "Static IP");
  }

  IO_WatchDogCounterEnable = false;
  FUOTA_WatchDogEnable = false;
  FuncIdOld = WAIT;

  //  start ticker activity every TIMER_MS
  hTimerCallback = xTimerCreate("Timer", TIMER_MS / portTICK_PERIOD_MS, pdTRUE, this, TimerCallback);
  xTimerStart(hTimerCallback, 0);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   WiFi events
///////////////////////////////////////////////////////////////////////////////////////////////////////

esp_err_t cIotModuleCom::WiFiEventHandler(void *pContext, system_event_t *event)
{
  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)pContext;

  //  limit log info
  if (!pIotModuleCom->Displayed1)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - WiFiEventHandler execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());
    pIotModuleCom->Displayed1 = true;
  }
  
  esp_err_t ret = ESP_OK;
  char ipString[15 + 1];

  switch (event->event_id)
  {
    case SYSTEM_EVENT_WIFI_READY:
      //  never seen
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_WIFI_READY", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_SCAN_DONE:
      //  SSID scan completed
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_SCAN_DONE", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_START:
      //  Station component started
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_START", __FILENAME__, __LINE__);
        //  get IotModule name
        char pModuleName[32 + 1];
        pIotModuleCom->IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));
        //  set IotModule host name
        ESP_ERROR_CHECK(tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_STA, pModuleName));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Host name : %s", __FILENAME__, __LINE__, pModuleName);
        //  create the UDP socket and begin receive loop in ComServerTask
        if (pIotModuleCom->hComServerTask != nullptr)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_START - xEventGroupSetBits", __FILENAME__, __LINE__);
          xEventGroupSetBits(pIotModuleCom->hEventGroup, 1);
        }
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - EVENT_STA_START - ComServerTask not available", __FILENAME__, __LINE__);
        }
      }
      break;

    case SYSTEM_EVENT_STA_STOP:
      //  happens only when FUOTA active and restarting with esp_restart()
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_STOP", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_CONNECTED:
      //  Station connected
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - connected to SSID : %s", __FILENAME__, __LINE__, pIotModuleCom->wifi_config.ap.ssid);
      wifi_ap_record_t AP_Info;
      ESP_ERROR_CHECK(esp_wifi_sta_get_ap_info(&AP_Info));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP Info - channel : %d", __FILENAME__, __LINE__, AP_Info.primary);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_CONNECTED - AP Info - RSSI : %d dBm", __FILENAME__, __LINE__, AP_Info.rssi);
      break;

    case SYSTEM_EVENT_STA_DISCONNECTED:
      //  Station disconnected
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_DISCONNECTED", __FILENAME__, __LINE__);
      //  do not connect when FUOTA active and restarting with esp_restart()
      if (!pIotModuleCom->FUOTA_WatchDogEnable)
      {
        ESP_ERROR_CHECK(esp_wifi_connect());
      }
      break;

    case SYSTEM_EVENT_STA_AUTHMODE_CHANGE:
      //  never seen - never tested
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_AUTHMODE_CHANGE", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_STA_GOT_IP:
      //  IP obtained by DHCP or Static IP set
      //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IPv4 : " IPSTR " - changed : %d", __FILENAME__, __LINE__, IP2STR(&event->event_info.got_ip.ip_info.ip), event->event_info.got_ip.ip_changed);
      ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_STA, &pIotModuleCom->tcpip_adapter_ip_info));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_GOT_IP - IP : " IPSTR " , GW : " IPSTR " , NM : " IPSTR ", IP changed : %d", __FILENAME__, __LINE__, IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.ip), IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.gw),
                          IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.netmask), event->event_info.got_ip.ip_changed);
      ESP_ERROR_CHECK(tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_STA, TCPIP_ADAPTER_DNS_MAIN, &pIotModuleCom->tcpip_adapter_dns_info));
      inet_ntop(AF_INET, &pIotModuleCom->tcpip_adapter_dns_info.ip, ipString, sizeof(ipString));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_GOT_IP - DNS - IP : %s", __FILENAME__, __LINE__, ipString);
      break;

    case SYSTEM_EVENT_STA_LOST_IP:
      //  happens after about 2 minuts from AP diconnection
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_STA_LOST_IP", __FILENAME__, __LINE__);
      break;

    case SYSTEM_EVENT_AP_START:
      {
        //  SoftAP component started
        //  get IotModule name
        char pModuleName[32 + 1];
        pIotModuleCom->IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));
        //  set IotModule host name
        ESP_ERROR_CHECK(tcpip_adapter_set_hostname(TCPIP_ADAPTER_IF_AP, pModuleName));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - Host name : %s", __FILENAME__, __LINE__, pModuleName);
        
        ESP_ERROR_CHECK(tcpip_adapter_get_ip_info(TCPIP_ADAPTER_IF_AP, &pIotModuleCom->tcpip_adapter_ip_info));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - IP : " IPSTR " , GW : " IPSTR " , NM : " IPSTR, __FILENAME__, __LINE__, IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.ip), IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.gw), IP2STR(&pIotModuleCom->tcpip_adapter_ip_info.netmask));
        ESP_ERROR_CHECK(tcpip_adapter_get_dns_info(TCPIP_ADAPTER_IF_AP, TCPIP_ADAPTER_DNS_MAIN, &pIotModuleCom->tcpip_adapter_dns_info));
        inet_ntop(AF_INET, &pIotModuleCom->tcpip_adapter_dns_info.ip, ipString, sizeof(ipString));
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - DNS - IP : %s", __FILENAME__, __LINE__, ipString);
        if (pIotModuleCom->hComServerTask != nullptr)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_START - xEventGroupSetBits", __FILENAME__, __LINE__);
          //  create the UDP socket and begin receive loop in ComServerTask
          xEventGroupSetBits(pIotModuleCom->hEventGroup, 1);
        }
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - SYSTEM_EVENT_AP_START - ComServerTask not available", __FILENAME__, __LINE__);
        }
      }
      break;

    case SYSTEM_EVENT_AP_STOP:
      //  it must never happens
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - EVENT_AP_STOP", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STACONNECTED:
      //  a Station is connected
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STACONNECTED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STADISCONNECTED:
      //  a Station is disconnected
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STADISCONNECTED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_PROBEREQRECVED:
      //  never seen - never tested
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_PROBEREQRECVED", __FILENAME__, __LINE__);
      break;
    case SYSTEM_EVENT_AP_STAIPASSIGNED:
      //  a Station get the IP
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STAIPASSIGNED", __FILENAME__, __LINE__);
      //=====================================================================================
      //  //  dhcp_search_ip_on_mac does not compile, compiler throws "undefined reference to `dhcp_search_ip_on_mac(unsigned char*, ip4_addr*)'"
      //  {
      //    wifi_sta_list_t station_list;
      //    esp_wifi_ap_get_sta_list(&station_list);
      //    wifi_sta_info_t *stations = (wifi_sta_info_t*)station_list.sta;
      //    dhcp_search_ip_on_mac(stations[0].mac , &event->event_info.got_ip.ip_info.ip);
      //  }
      //  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "WiFiEventHandler - IPv4 : %s - changed : %d", ip4addr_ntoa(&event->event_info.got_ip.ip_info.ip), event->event_info.got_ip.ip_changed);
      //=====================================================================================
      break;
    case SYSTEM_EVENT_AP_STA_GOT_IP6:
      //  not implemented
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STA_GOT_IP6", __FILENAME__, __LINE__);
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_AP_STA_GOT_IP6 - IPv6 : " IPV6STR, __FILENAME__, __LINE__, IPV62STR(event->event_info.got_ip6.ip6_info.ip));
      break;
    default:
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - EVENT_NOT_HANDLED - Event : %d", __FILENAME__, __LINE__, event->event_id);
      break;
  }
  //  feed event to arduino loop()
  if (pIotModuleCom->WiFiEventHandlerCallback != nullptr) pIotModuleCom->WiFiEventHandlerCallback(pIotModuleCom->pIotModuleContext, event);
  return ret;
}


///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   UDP Communication server task
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleCom::ComServerTask(void *pvParameters)
{
  struct sockaddr_in SockAddr;
  char pComBuf[UDP_COM_BUF_LEN];
  int ComBufLen;
#ifdef IOTMODULE_DEBUG
  int64_t Begin_uSec, End_uSec;
#endif
  bool StayInLoop = true;

  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)pvParameters;

  do
  {
    //  start when station or soft AP initialization is done
    EventBits_t WaitBits = xEventGroupWaitBits(pIotModuleCom->hEventGroup, 1, pdTRUE, pdFALSE, portMAX_DELAY );
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - xEventGroupSetBits : %d", __FILENAME__, __LINE__, WaitBits);
    //  if at least one bit set the event is received else skip the code body because timeout elapsed (portMAX_DELAY means about 40 days) and loop again
    if (WaitBits != 0)
    {
      //  create UDP socket
      if ((pIotModuleCom->hIotHALSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) != -1)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - UDP socket created", __FILENAME__, __LINE__);

        //  not necessary
        //  SO_BROADCAST
        //  Set or get the broadcast flag. When enabled, datagram sockets are allowed to send packets to a broadcast address. This option has no effect on stream-oriented sockets.
        //  In this case we don't send broadcast, we only receive them so is not necessary
        //  int opt_flag = 1;
        //  ESP_ERROR_CHECK(setsockopt(hIotHALSocket, SOL_SOCKET, SO_BROADCAST, &opt_flag, sizeof(opt_flag)));

        SockAddr.sin_family = AF_INET;
        SockAddr.sin_port = htons(IOTMODULE_UDP_PORT_NUMBER);
        SockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
        socklen_t SockAddrLen;

        //  set socket address length
        SockAddrLen = sizeof(struct sockaddr_in);
        // bind socket
        if (bind(pIotModuleCom->hIotHALSocket, (struct sockaddr *)&SockAddr, SockAddrLen) == 0)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - bind active, falling into recvfrom loop", __FILENAME__, __LINE__);
#ifdef IOTMODULE_DEBUG
          //  start perf measure
          Begin_uSec = esp_timer_get_time();
#endif
          do
          {
            //  get the last inputs before wait for new RPC
            pIotModuleCom->IotModuleApp.GetInputValues();
#ifdef IOTMODULE_DEBUG
            //  stop measure active phase
            End_uSec = esp_timer_get_time();
            //  set active phase value in generic
            pIotModuleCom->IotModuleApp.GenericWriteFloat(3, End_uSec - Begin_uSec);
            //  start measure inactive phase
            Begin_uSec = End_uSec;
#endif
            //  set socket address length
            SockAddrLen = sizeof(struct sockaddr_in);
            //  received 0 bytes (when used nmap -sU) or 1 or more bytes during normal operation
            if ((ComBufLen = recvfrom(pIotModuleCom->hIotHALSocket, pComBuf, UDP_COM_BUF_LEN, 0, (struct sockaddr *) &SockAddr, &SockAddrLen)) != -1)
            {
#ifdef IOTMODULE_DEBUG              
              //  stop measure inactive phase
              End_uSec = esp_timer_get_time();
              //  set inactive phase value in generic
              pIotModuleCom->IotModuleApp.GenericWriteFloat(9, End_uSec - Begin_uSec);
              //  start measure active phase
              Begin_uSec = End_uSec;
#endif
              //  enable and reset watchdog
              pIotModuleCom->IO_WatchDogCounterEnable = true;
              pIotModuleCom->IO_WatchDogCounter = 0;

              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received packet from %s:%d", __FILENAME__, __LINE__, inet_ntoa(SockAddr.sin_addr), ntohs(SockAddr.sin_port));
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received data - len : %d", __FILENAME__, __LINE__, ComBufLen);
              pIotModuleCom->PeerIP_Addr = SockAddr.sin_addr;

              //  RPC data, decode request and encode reply
              pIotModuleCom->NetModuleRPC((void*)pComBuf, &ComBufLen);

              //  if replay requested
              if (ComBufLen != 0)
              {
                //  if replay data can be transmitted
                if (ComBufLen <= sizeof(pComBuf))
                {
                  int Len = sendto(pIotModuleCom->hIotHALSocket, pComBuf, ComBufLen, 0, (struct sockaddr *)&SockAddr, SockAddrLen);
                  //  sent ok
                  if ((Len != -1) && (Len == ComBufLen))
                  {
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - sendto : sent packet to %s:%d", __FILENAME__, __LINE__, inet_ntoa(SockAddr.sin_addr), ntohs(SockAddr.sin_port));
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - sendto : data - len : %d", __FILENAME__, __LINE__, ComBufLen);

                    //  set the last ouput values received from RPC
                    pIotModuleCom->IotModuleApp.SetOutputValues(false);
                    StayInLoop = true;
                  }
                  //  ENOMEM
                  else if ((Len == -1) && (errno == ENOMEM))
                  {
                    //  https://docs.espressif.com/projects/esp-idf/en/latest/api-guides/lwip.html
                    //  Calling send() or sendto() repeatedly on a UDP socket may eventually fail with errno equal to ENOMEM. 
                    //  This is a limitation of buffer sizes in the lower layer network interface drivers. 
                    //  If all driver transmit buffers are full then UDP transmission will fail. 
                    //  Applications sending a high volume of UDP datagrams who don’t wish for any to be dropped by the sender should check for this error code and re-send the datagram after a short delay. 
                    //  Increasing the number of TX buffers in the Wi-Fi or Ethernet project configuration (as applicable) may also help.
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto ENOMEM error", __FILENAME__, __LINE__);
                    StayInLoop = true;
                  }
                  //  other error
                  else
                  {
                    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto error : %d", __FILENAME__, __LINE__, errno);
                    StayInLoop = false;
                  }
                }
                //  data size exeed buffer size
                else
                {
                  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto - data size exeed buffer size", __FILENAME__, __LINE__);
                  StayInLoop = false;
                }
              }
              //  data not recognized, data size equal zero
              else
              {
                ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - sendto - 0 data size", __FILENAME__, __LINE__);
                StayInLoop = true;
              }
            }
            else
            //  recvfrom error
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - recvfrom error : %d", __FILENAME__, __LINE__, errno);
              StayInLoop = false;
            }
          }
          while (StayInLoop);
        }
        //  bind error
        else
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - socket bind error : %d", __FILENAME__, __LINE__, errno);
          StayInLoop = false;
        }
        shutdown(pIotModuleCom->hIotHALSocket, 0);
        close(pIotModuleCom->hIotHALSocket);
      }
      //  error creating socket
      else
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - UDP socket creation error : %d", __FILENAME__, __LINE__, errno);
        StayInLoop = false;
      }
    }
  }
  while (StayInLoop);

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - FATAL ERROR! communication task deleted!", __FILENAME__, __LINE__);
  TaskHandle_t hTmp = pIotModuleCom->hComServerTask;
  pIotModuleCom->hComServerTask = nullptr;
  vTaskDelete(hTmp);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                   Remote Procedure Call
//                  Set data structures to / from  communication data buffer
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleCom::NetModuleRPC(const void *pComBuf, int *pIotHALDataLen)
{
  tIotHALData *pIotHALData = (tIotHALData *)pComBuf;

  //  Received at least Protocol and ProtocolVer?
  if (*pIotHALDataLen >= (STRUCTMEMBERSIZE(sHeader, Protocol) + STRUCTMEMBERSIZE(sHeader, ProtocolVer)))
  {
    //  Protocol and ProtocolVer?
    if (memcmp(pIotHALData->Header.Protocol, IOTMODULE_THEREMINO_IOT_PROTOCOL, 6) == 0 && pIotHALData->Header.ProtocolVer == IOTMODULE_THEREMINO_IOT_PROTOCOL_VER)
    {
      //  FirmwareVer ignored for now
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.Who : %c", __FILENAME__, __LINE__, pIotHALData->Header.Who);
      //  Who?
      if (pIotHALData->Header.Who == 'H')
      {
        //  Command?
        switch (pIotHALData->Header.FuncId)
        {
          case SCAN:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : SCAN", __FILENAME__, __LINE__);

              //  Scan replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              memcpy(&pIotHALData->Body.ScanReply.MAC, WiFiMAC, sizeof(pIotHALData->Body.ScanReply.MAC));
              IotModuleApp.RPCGetModuleName(&pIotHALData->Body.ScanReply.ModuleNameSize, pIotHALData->Body.ScanReply.ModuleName);
              pIotHALData->Body.ScanReply.PinoutType = IOTMODULE_PINOUT_TYPE;
              pIotHALData->Body.ScanReply.ProgStatus = IotModuleApp.GetProgStatus();
              pIotHALData->Body.ScanReply.TotalPinsNumber = IotModuleApp.RPCGetTotalPinsNumber();
              for (uint8_t i = 0; i < IotModuleApp.RPCGetTotalPinsNumber(); i++)
              {
                pIotHALData->Body.ScanReply.PinName[i] = IotModuleApp.RPCGetPinName(i);
              }
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.ScanReply);
            }
            break;

          case PROG:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : PROG", __FILENAME__, __LINE__);

              //  Prog request RPC
              IotModuleApp.RPCSetAllPinsParams(pIotHALData->Body.ProgReq.TotalPinsNumber, pIotHALData->Body.ProgReq.PinType);

              //  Prog replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.ProgReply.ProgStatus = IotModuleApp.GetProgStatus();
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.ProgReply);
            }
            break;

          case GETSINGLE:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : GETSINGLE", __FILENAME__, __LINE__);

              //  GetPinTypes request RPC
              uint8_t tmp1 = pIotHALData->Body.GetSingleReq.PinIndex;

              //  GetPinTypes replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.GetSingleReply.PinName = IotModuleApp.RPCGetPinName(tmp1);
              IotModuleApp.RPCGetDynamicPinTypes(tmp1, &pIotHALData->Body.GetSingleReply.TotalPinTypes, pIotHALData->Body.GetSingleReply.PinType);
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.GetSingleReply);
            }
            break;

          case XCHG:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : XCHG", __FILENAME__, __LINE__);

              if (IotModuleApp.GetProgStatus() == 'Y')
              {
                //  Xchg request RPC (outputs)
                uint8_t tmp1 = pIotHALData->Body.XchgReq.TotalOutValues;
                if (IotModuleApp.GetOutputPinsNumber() == tmp1)
                {
                  for (uint8_t i = 0; i < tmp1; i++)
                  {
                    IotModuleApp.RPCSetOutputXchgValues(i, &pIotHALData->Body.XchgReq.OutValue[i]);
                  }

                  //  Xchg replay RPC (inputs)
                  //  pIotHALData->Header.Command = <the same>
                  pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                  pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                  pIotHALData->Header.Who = 'M';
                  //  speed it up, do not call GetProgrammed, test done before
                  pIotHALData->Body.XchgReply.ProgStatus = 'Y';
                  tmp1 = IotModuleApp.GetInputPinsNumber();
                  pIotHALData->Body.XchgReply.TotalInValues = tmp1;
                  for (uint8_t i = 0; i < tmp1; i++)
                  {
                    IotModuleApp.RPCGetInputXchgValues(i, &pIotHALData->Body.XchgReply.InValue[i]);
                  }
                  *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
                }
                else
                {
                  ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Header.FuncId : XCHG - output pin number mismatch - req : %d - prog : %d", __FILENAME__, __LINE__, tmp1, IotModuleApp.GetOutputPinsNumber());

                  //  Xchg replay RPC (inputs)
                  //  pIotHALData->Header.Command = <the same>
                  pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                  pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                  pIotHALData->Header.Who = 'M';
                  //  todoo ProgStatus = 'N'
                  pIotHALData->Body.XchgReply.ProgStatus = 'N';
                  pIotHALData->Body.XchgReply.TotalInValues = 0;
                  *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
                }
              }
              else
              {
                ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - Header.FuncId : XCHG - called xchg but module not programmed", __FILENAME__, __LINE__);

                //  Xchg replay RPC (inputs)
                //  pIotHALData->Header.Command = <the same>
                pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
                pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
                pIotHALData->Header.Who = 'M';
                //  speed it up, do not call GetProgrammed, test done before
                pIotHALData->Body.XchgReply.ProgStatus = 'N';
                pIotHALData->Body.XchgReply.TotalInValues = 0;
                *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.XchgReply);
              }
            }
            break;

          case NAME:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : NAME", __FILENAME__, __LINE__);

              //  Name request RPC
              IotModuleApp.RPCSetModuleName(pIotHALData->Body.NameReq.ModuleNameSize, pIotHALData->Body.NameReq.ModuleName);

              //  Name replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              IotModuleApp.RPCGetModuleName(&pIotHALData->Body.NameReply.ModuleNameSize, pIotHALData->Body.NameReply.ModuleName);
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.NameReply);
            }
            break;

          case EXEC:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : EXEC", __FILENAME__, __LINE__);

              if (IOTMODULE_FUOTA_ENABLE == 1)
              {
                //  stop app dangerous activities
                IotModuleApp.PreFuota();
                //  start Firmware Upgrade On The Air
                Fuota();
              }
            }
            break;

          case PWM:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : PWM", __FILENAME__, __LINE__);

              //  Pwm request RPC
              uint8_t PinIndex = pIotHALData->Body.PwmReq.PinIndex;
              float Freq = pIotHALData->Body.PwmReq.Freq;
              IotModuleApp.RPCPwmSetFrequency(PinIndex, Freq);

              //  Pwm replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.PwmReply.ReceivedOK = true;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.PwmReply);
            }
            break;
          //  todoo stepper
          case STEP:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : STEP", __FILENAME__, __LINE__);

              //  Step request RPC
              uint8_t PinIndex = pIotHALData->Body.StepReq.PinIndex;
              uint32_t SpeedHz = pIotHALData->Body.StepReq.SpeedHz;
              uint16_t AccHz_mSec = pIotHALData->Body.StepReq.AccHz_mSec;
              bool RetVal = IotModuleApp.StepperConfigSpeedAcc(PinIndex, SpeedHz, AccHz_mSec);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - PinIndex : %d, SpeedHz : %d , AccHz_mSec : %d", __FILENAME__, __LINE__, PinIndex, SpeedHz, AccHz_mSec);
              
              //  Step replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.StepReply.ReceivedOK = RetVal;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.StepReply);
            }
            break;
          case ADC24:
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - Header.FuncId : ADC24", __FILENAME__, __LINE__);

              //  ADC24 request RPC
              uint16_t Sps = pIotHALData->Body.Adc24Req.Sps;
              bool MultipleSetups = pIotHALData->Body.Adc24Req.MultipleSetups;
              bool FiltersSingleCycle = pIotHALData->Body.Adc24Req.FiltersSingleCycle;
              uint8_t PreFilterType = pIotHALData->Body.Adc24Req.PreFilterType;
              uint8_t PostFilterType = pIotHALData->Body.Adc24Req.PostFilterType;
              uint16_t ActiveChannels = pIotHALData->Body.Adc24Req.ActiveChannels;
              uint16_t BiasedChannels = pIotHALData->Body.Adc24Req.BiasedChannels;
              uint16_t DiffChannels = pIotHALData->Body.Adc24Req.DiffChannels;
              uint16_t PseudoChannels = pIotHALData->Body.Adc24Req.PseudoChannels;
              uint8_t ChannelsGains0 = pIotHALData->Body.Adc24Req.ChannelsGains[0];
              uint8_t ChannelsGains1 = pIotHALData->Body.Adc24Req.ChannelsGains[1];
              uint8_t ChannelsGains2 = pIotHALData->Body.Adc24Req.ChannelsGains[2];
              IotModuleApp.ADC24SetParam(Sps, MultipleSetups, FiltersSingleCycle, PreFilterType,
                                         PostFilterType, ActiveChannels, BiasedChannels, DiffChannels,
                                         PseudoChannels, ChannelsGains0, ChannelsGains1, ChannelsGains2);
              //  ADC24 replay RPC
              //  pIotHALData->Header.Command = <the same>
              pIotHALData->Header.ProtocolVer = IOTMODULE_THEREMINO_IOT_PROTOCOL_VER;
              pIotHALData->Header.FirmwareVer = IOTMODULE_FIRMWARE_VER;
              pIotHALData->Header.Who = 'M';
              pIotHALData->Body.Adc24Reply.ReceivedOK = true;
              *pIotHALDataLen = sizeof(pIotHALData->Header) + sizeof(pIotHALData->Body.Adc24Reply);
            }
            break;
          default:
            //  FUNC_ID not recognized
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - FUNC_ID not recognized", __FILENAME__, __LINE__);
            //  do not replay
            *pIotHALDataLen = 0;
            break;
        }
        if (FuncIdOld != pIotHALData->Header.FuncId)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - IotHAL IP : %s - FUNC_ID : %d executed ", __FILENAME__, __LINE__, inet_ntoa(PeerIP_Addr), pIotHALData->Header.FuncId);
          FuncIdOld = pIotHALData->Header.FuncId;
        }
      }
      else
      {
        //  sender is not HAL
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sender is not HAL", __FILENAME__, __LINE__);
        //  do not replay
        *pIotHALDataLen = 0;
      }
    }
    else
    {
      //  protocol not recognized
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - protocol not recognized", __FILENAME__, __LINE__);
      //  do not replay
      *pIotHALDataLen = 0;
    }
  }
  else
  {
    //  data malformed
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - data malformed", __FILENAME__, __LINE__);
    //  do not replay
    *pIotHALDataLen = 0;
  }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                         Firmware Upgrade On The Air (FUOTA) fo Arduino IDE
///////////////////////////////////////////////////////////////////////////////////////////////////////

void cIotModuleCom::Fuota(void)
{
  //  MDNS TCP port
  int port = 3232;
  char hostName[32 + 1 + 12];
  mdns_txt_item_t arduTxtData[4] =
  {
    {(char*)"board"         , (char*)ARDUINO_VARIANT},
    {(char*)"tcp_check"     , (char*)"no"},
    {(char*)"ssh_upload"    , (char*)"no"},
    {(char*)"auth_upload"   , (char*)"no"}
  };

  char pModuleName[32 + 1];
  IotModuleApp.GetModuleName(pModuleName, sizeof(pModuleName));

  //  restart the module after FUOTA_WATCHDOG_DELAY_MS if FUOTA is never done
  FUOTA_WatchDogCounter = 0;
  FUOTA_WatchDogEnable = true;

  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA starting, waiting %d seconds before automatic reboot", __FILENAME__, __LINE__, FUOTA_WATCHDOG_DELAY_MS / TIMER_MS / 100);

  sprintf(hostName, "%s at %02X:%02X:%02X:%02X:%02X:%02X", pModuleName, WiFiMAC[0], WiFiMAC[1], WiFiMAC[2], WiFiMAC[3], WiFiMAC[4], WiFiMAC[5]);
  ESP_ERROR_CHECK(mdns_init());
  ESP_ERROR_CHECK(mdns_hostname_set(hostName));
  ESP_ERROR_CHECK(mdns_service_add(NULL, "_arduino", "_tcp", port, arduTxtData, 4));
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MDNS executed : %s", __FILENAME__, __LINE__, hostName);

  struct sockaddr_in ArduinoIDE_UdpAddr;
  char pArduinoIDE_UdpData[UDP_COM_BUF_LEN];
  unsigned int ArduinoIDE_UdpDataLen;
  int hArduinoIDE_UdpSocket;

  // create UDP socket
  if ((hArduinoIDE_UdpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) >= 0)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - UDP socket created", __FILENAME__, __LINE__);

    //  SO_BROADCAST
    //  Set or get the broadcast flag. When enabled, datagram sockets are allowed to send packets to a broadcast address. This option has no effect on stream-oriented sockets.
    //  In this case we don't send broadcast, we only receive them so it is not necessary
    //  int opt_flag = 1;
    //  ESP_ERROR_CHECK(setsockopt(hArduinoIDE_UdpSocket, SOL_SOCKET, SO_BROADCAST, &opt_flag, sizeof(opt_flag)));

    ArduinoIDE_UdpAddr.sin_family = AF_INET;
    ArduinoIDE_UdpAddr.sin_port = htons(ARDUINO_IDE_UDP_PORT_NUMBER);
    ArduinoIDE_UdpAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    // bind socket
    if (bind(hArduinoIDE_UdpSocket, (struct sockaddr *)&ArduinoIDE_UdpAddr, sizeof(ArduinoIDE_UdpAddr)) >= 0)
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - bind active, falling into recvfrom loop", __FILENAME__, __LINE__);

      if ((ArduinoIDE_UdpDataLen = recvfrom(hArduinoIDE_UdpSocket, pArduinoIDE_UdpData, UDP_COM_BUF_LEN, 0, (struct sockaddr *)&ArduinoIDE_UdpAddr, &ArduinoIDE_UdpDataLen)) <= 0)
      {
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - recvfrom error : %d", __FILENAME__, __LINE__, errno);
      }

      ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA data transfer started", __FILENAME__, __LINE__);

      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received packet from %s:%d", __FILENAME__, __LINE__, inet_ntoa(ArduinoIDE_UdpAddr.sin_addr), ntohs(ArduinoIDE_UdpAddr.sin_port));
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received raw data : %s - len %d", __FILENAME__, __LINE__ , pArduinoIDE_UdpData, ArduinoIDE_UdpDataLen);

      char delimiter = ' ';
      std::string tobesplit(pArduinoIDE_UdpData);
      std::vector<std::string> tokens;
      std::string token;
      std::istringstream tokenStream(tobesplit);
      while (std::getline(tokenStream, token, delimiter))
      {
        tokens.push_back(token);
      }

      if ((tokens.size() == 4) && (atoi(tokens[0].c_str()) == 0))
      {
        int OTA_TcpPort = atoi(tokens[1].c_str());
        int OTA_AppSize = atoi(tokens[2].c_str());
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - received data - OTA_TcpPort : %d - OTA_AppSize %d - MD5 : %s", __FILENAME__, __LINE__,  OTA_TcpPort, OTA_AppSize, tokens[3].c_str());

        //  call esp_ota_begin here, before TCP initialization, to avoid strange behaviour in TCP session
        //  esp_ota_begin call "freeze" communications for about 3 seconds
        //  todo : bug? retry with newer arduino esp32 board's library version
        const esp_partition_t *update_partition = esp_ota_get_next_update_partition(NULL);
        esp_ota_handle_t update_handle;
        ESP_ERROR_CHECK(esp_ota_begin(update_partition, OTA_SIZE_UNKNOWN, &update_handle));

        sprintf(pArduinoIDE_UdpData, "%s", "OK");
        ArduinoIDE_UdpDataLen = strlen(pArduinoIDE_UdpData);
        if (sendto(hArduinoIDE_UdpSocket, pArduinoIDE_UdpData, ArduinoIDE_UdpDataLen, 0, (struct sockaddr *)&ArduinoIDE_UdpAddr, sizeof(ArduinoIDE_UdpAddr)) < 0)
        {
          ESP_ERROR_CHECK(esp_ota_end(update_handle));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - sendto error : %d", __FILENAME__, __LINE__, errno);
        }

        struct sockaddr_in ArduinoIDE_TcpAddr;
        int hArduinoIDE_TcpSocket;
        char pArduinoIDE_TcpData[FUOTA_TCP_DATA_BUF_LEN];
        unsigned int ArduinoIDE_TcpDataLen;

        //  create TCP socket
        if ((hArduinoIDE_TcpSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) >= 0)
        {
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket created", __FILENAME__, __LINE__);

          ArduinoIDE_TcpAddr.sin_addr.s_addr = ArduinoIDE_UdpAddr.sin_addr.s_addr;
          ArduinoIDE_TcpAddr.sin_family = AF_INET;
          ArduinoIDE_TcpAddr.sin_port = htons(OTA_TcpPort);
          if (connect(hArduinoIDE_TcpSocket, (struct sockaddr *)&ArduinoIDE_TcpAddr, sizeof(ArduinoIDE_TcpAddr)) == 0)
          {
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket connected, starting FUOTA", __FILENAME__, __LINE__);

            unsigned int OTA_AppSizeCounter = 0;
            unsigned int OTA_AppChunk;
            MD5Builder MD5;
            MD5.begin();

            do
            {
              ArduinoIDE_TcpDataLen = recv(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, FUOTA_TCP_DATA_BUF_LEN, 0);
              ESP_ERROR_CHECK(esp_ota_write(update_handle, (const void *)pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen));
              OTA_AppSizeCounter += ArduinoIDE_TcpDataLen;
              OTA_AppChunk = ArduinoIDE_TcpDataLen;
              MD5.add(reinterpret_cast<uint8_t*>(pArduinoIDE_TcpData), ArduinoIDE_TcpDataLen);
              sprintf(pArduinoIDE_TcpData, "%u", ArduinoIDE_TcpDataLen);
              ArduinoIDE_TcpDataLen = strlen(pArduinoIDE_TcpData);
              ArduinoIDE_TcpDataLen = send(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen, 0);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - chunk %u - total %u", __FILENAME__, __LINE__, OTA_AppChunk, OTA_AppSizeCounter);
            }
            while ((ArduinoIDE_TcpDataLen > 0) && (OTA_AppSize != OTA_AppSizeCounter));

            sprintf(pArduinoIDE_TcpData, "%s", "OK");
            ArduinoIDE_TcpDataLen = strlen(pArduinoIDE_TcpData);
            ArduinoIDE_TcpDataLen = send(hArduinoIDE_TcpSocket, pArduinoIDE_TcpData, ArduinoIDE_TcpDataLen, 0);
            shutdown(hArduinoIDE_TcpSocket, 0);
            close(hArduinoIDE_TcpSocket);
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - TCP socket closed", __FILENAME__, __LINE__);

            ESP_ERROR_CHECK(esp_ota_end(update_handle));
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - esp_ota_end", __FILENAME__, __LINE__);

            FUOTA_WatchDogEnable = false;

            MD5.calculate();
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MD5s : %s", __FILENAME__, __LINE__, tokens[3].c_str());
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - MD5d : %s", __FILENAME__, __LINE__, MD5.toString().c_str());
            char pMD5s[33], pMD5d[33];
            unsigned int pMD5sl, pMD5dl;
            strncpy(pMD5s, tokens[3].c_str(), 32);
            pMD5s[32] = 0;
            strcpy(pMD5d, MD5.toString().c_str());
            pMD5sl = strlen(pMD5s);
            pMD5dl = strlen(pMD5d);
            int r = strcmp(pMD5s, pMD5d);
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - pMD5sl : %d - pMD5dl : %d - strcmp : %d", __FILENAME__, __LINE__, pMD5sl, pMD5dl, r);
            if (r == 0)
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA data transfer completed", __FILENAME__, __LINE__);
              ESP_ERROR_CHECK(esp_ota_set_boot_partition(update_partition));
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - esp_ota_set_boot_partition", __FILENAME__, __LINE__);
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA set new boot partition", __FILENAME__, __LINE__);
            }
            else
            {
              ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA data transfer failed", __FILENAME__, __LINE__);
            }
          }
          else
          {
            ESP_ERROR_CHECK(esp_ota_end(update_handle));
            ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - TCP socket connection error : %d", __FILENAME__, __LINE__, errno);
          }
        }
        else
        {
          ESP_ERROR_CHECK(esp_ota_end(update_handle));
          ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - TCP socket creation error : %d", __FILENAME__, __LINE__, errno);
        }
      }
      shutdown(hArduinoIDE_UdpSocket, 0);
      close(hArduinoIDE_UdpSocket);
    }
    else
    {
      ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - socket bind error : %d", __FILENAME__, __LINE__, errno);
      shutdown(hArduinoIDE_UdpSocket, 0);
      close(hArduinoIDE_UdpSocket);
      hArduinoIDE_UdpSocket = 0;
    }
  }
  else
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - UDP socket creation error : %d", __FILENAME__, __LINE__, errno);
  }

  ESP_ERROR_CHECK(mdns_service_remove_all());
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - mdns_service_remove_all", __FILENAME__, __LINE__);
  //  this delay is needed between mdns_service_remove_all() and mdns_free()
  //  todo : bug? retry with newer arduino esp32 board's library version
  //  2020/04/04 seems ok now
  //  vTaskDelay(1000 / portTICK_PERIOD_MS);
  mdns_free();
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_VERBOSE, pTAG, "%18s %4d - mdns_free", __FILENAME__, __LINE__);
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA successfully completed, esp_restart in 5 sec", __FILENAME__, __LINE__);
  vTaskDelay(5000 / portTICK_PERIOD_MS);
  esp_restart();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////
//                                  TimerCallback called every x mSec
///////////////////////////////////////////////////////////////////////////////////////////////////////

void IRAM_ATTR cIotModuleCom::TimerCallback(TimerHandle_t pxTimer)
{
  cIotModuleCom *pIotModuleCom = (cIotModuleCom *)pvTimerGetTimerID( pxTimer );

  if (!pIotModuleCom->Displayed2)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - TimerCallback execution core : %d", __FILENAME__, __LINE__, xPortGetCoreID());
    pIotModuleCom->Displayed2 = true;
  }

  if (IOTMODULE_OUTPUT_WATCHDOG_ENABLE == 1 && pIotModuleCom->IO_WatchDogCounterEnable == true)
  {
    //  if delay not triggered yet
    if (pIotModuleCom->IO_WatchDogCounter <= IO_WATCHDOG_DELAY_MS / TIMER_MS)
    {
      //  if delay triggered
      if (pIotModuleCom->IO_WatchDogCounter == IO_WATCHDOG_DELAY_MS / TIMER_MS)
      {
        //  reset output values
        pIotModuleCom->IotModuleApp.SetOutputValues(true);
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - output watchdog triggered, output values have been reset", __FILENAME__, __LINE__);
      }
      pIotModuleCom->IO_WatchDogCounter++;
    }
  }
  if (pIotModuleCom->FUOTA_WatchDogEnable)
  {
    //  if delay not triggered yet
    if (pIotModuleCom->FUOTA_WatchDogCounter <= FUOTA_WATCHDOG_DELAY_MS / TIMER_MS)
    {
      //  if delay triggered
      if (pIotModuleCom->FUOTA_WatchDogCounter == FUOTA_WATCHDOG_DELAY_MS / TIMER_MS)
      {
        //  log reboot
        ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - FUOTA wait period expired, esp_restart in 5 sec", __FILENAME__, __LINE__);
        //  delay 5 Sec
        vTaskDelay(5000 / portTICK_PERIOD_MS);
        //  reboot system
        esp_restart();
      }
      pIotModuleCom->FUOTA_WatchDogCounter++;
    }
  }
  //  Stepper interpolator
  pIotModuleCom->IotModuleApp.TimedInterpolate();
}
