//================================================================================================
//                                  cIotModuleServo body
//================================================================================================

void cIotModuleServo::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  
  //  pointer to App object used to address Generic* functions
  this->pIotModuleApp = pIotModuleApp;
    
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    Data.Params[Channel].Attached = false;
  }
}
//  called from cIotModuleApp
bool cIotModuleServo::Attach(const uint8_t LedcChannel, const uint8_t LedcTimer, const gpio_num_t Name)
{
  if (!Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {    
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - LedcChannel : %d, LedcTimer : %d, Name : %d", __FILENAME__, __LINE__, LedcChannel, LedcTimer, Name);
    
    //  prepare and then apply the ledc timer configuration
    ledc_timer.speed_mode      = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    ledc_timer.timer_num       = (ledc_timer_t)(LedcTimer % 4);
    ledc_timer.duty_resolution = LEDC_DUTY;
    ledc_timer.freq_hz         = SPEED;
    ESP_ERROR_CHECK(ledc_timer_config(&ledc_timer));

    ESP_ERROR_CHECK(ledc_timer_resume(LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(LedcTimer % 4)));

    //  prepare and then apply the ledc channel configuration, it sets implicitly gpio_num GPIO as output we use for servo signal
    ledc_channel.speed_mode = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    //  odd channels not used
    ledc_channel.channel    = (ledc_channel_t)(LedcChannel % 8);
    ledc_channel.timer_sel  = (ledc_timer_t)(LedcTimer % 4);
    ledc_channel.intr_type  = LEDC_INTR_DISABLE;
    ledc_channel.gpio_num   = Name;
    ledc_channel.duty       = 1 << (LEDC_DUTY - 1);
    //  **************** set hpoint to 1 to get step GPIO low always after config *******************
    //  ********** if hpoint set to 0 step GPIO goes high after first use of ledc channel ***********
    ledc_channel.hpoint     = 1;
    ESP_ERROR_CHECK(ledc_channel_config(&ledc_channel));    
    Data.Params[LedcChannel].Attached = true;
  }
  return true;

}
//  called from cIotModuleApp
bool cIotModuleServo::Detach(const uint8_t LedcChannel)
{
  if (Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(ledc_stop(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8), 0));

    Data.Params[LedcChannel].Attached = false;
  }
  return true;
}
//  fast
bool IRAM_ATTR cIotModuleServo::SetDuty(const uint8_t LedcChannel, float Duty)
{
  if (Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    //  esp_err_t ledc_set_duty_and_update(ledc_mode_t speed_mode, ledc_channel_t channel, uint32_t duty, uint32_t hpoint)
    //  ESP_ERROR_CHECK(ledc_set_duty_and_update(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8), (Duty / 20000.0f) * (float)(1 << LEDC_DUTY) , 0));

    //  ledc_set_duty(ledc_mode_t speed_mode, ledc_channel_t channel, uint32_t duty)
    ESP_ERROR_CHECK(ledc_set_duty(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8), (Duty / 20000.0f) * (float)(1 << LEDC_DUTY)));

    //  esp_err_t ledc_update_duty(ledc_mode_t speed_mode, ledc_channel_t channel)
    ESP_ERROR_CHECK(ledc_update_duty(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8)));
    
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Servo Duty : %f", __FILENAME__, __LINE__, Duty);
  }
  return true;
}

//================================================================================================
//                                  cIotModulePWM body
//================================================================================================

void cIotModulePWM::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  
  //  pointer to App object used to address Generic* functions
  this->pIotModuleApp = pIotModuleApp;
    
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    Data.Params[Channel].Attached = false;
  }
}
//  called from cIotModuleApp
bool cIotModulePWM::Attach(const uint8_t LedcChannel, const uint8_t LedcTimer, const gpio_num_t Name)
{
  if (!Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    //  prepare and then apply the ledc timer configuration
    ledc_timer.speed_mode      = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    ledc_timer.timer_num       = (ledc_timer_t)(LedcTimer % 4);
    ledc_timer.duty_resolution = LEDC_DUTY;
    ledc_timer.freq_hz         = SPEED;
    ESP_ERROR_CHECK(ledc_timer_config(&ledc_timer));

    ESP_ERROR_CHECK(ledc_timer_resume(LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)(LedcTimer % 4)));

    //  prepare and then apply the ledc channel configuration, it sets implicitly gpio_num GPIO as output we use for PWM signal
    ledc_channel.speed_mode = LedcTimer / 4 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    //  odd channels not used
    ledc_channel.channel    = (ledc_channel_t)(LedcChannel % 8);
    ledc_channel.timer_sel  = (ledc_timer_t)(LedcTimer % 4);
    ledc_channel.intr_type  = LEDC_INTR_DISABLE;
    ledc_channel.gpio_num   = Name;
    ledc_channel.duty       = 1 << (LEDC_DUTY - 1);
    //  **************** set hpoint to 1 to get step GPIO low always after config *******************
    //  ********** if hpoint set to 0 step GPIO goes high after first use of ledc channel ***********
    ledc_channel.hpoint     = 1;
    ESP_ERROR_CHECK(ledc_channel_config(&ledc_channel));
    
    Data.Params[LedcChannel].Name = Name;
    Data.Params[LedcChannel].Attached = true;
  }
  return true;

}
//  called from cIotModuleApp
bool cIotModulePWM::Detach(const uint8_t LedcChannel)
{
  if (Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(ledc_stop(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8), 0));
    
    Data.Params[LedcChannel].Attached = false;
  }
  return true;
}
//  fast, called from cIotModuleApp
bool IRAM_ATTR cIotModulePWM::SetDuty(const uint8_t LedcChannel, float Duty)
{
  if (Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    //  ledc_set_duty(ledc_mode_t speed_mode, ledc_channel_t channel, uint32_t duty)
    ESP_ERROR_CHECK(ledc_set_duty(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8), Duty  * (float)Data.Params[LedcChannel].MaxWidth));

    //  esp_err_t ledc_update_duty(ledc_mode_t speed_mode, ledc_channel_t channel)
    ESP_ERROR_CHECK(ledc_update_duty(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_channel_t)(LedcChannel % 8)));
    
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - PWM Duty : %f", __FILENAME__, __LINE__, Duty);
  }
  return true;
}
bool cIotModulePWM::SetFreq(const uint8_t LedcChannel, const float Freq)
{
  
  if (Data.Params[LedcChannel % MAX_CHANNELS].Attached)
  {
    uint64_t _Freq = Freq;
    if (Freq < 1.0f) _Freq = 1;
    if (Freq > 40000000.0f) _Freq = 40000000;
    //  calculate the resolution
    Data.Params[LedcChannel].MaxWidth = 1 << 20;
    uint8_t ResBits = 20;
    while (ResBits >= 2)
    {
      if ((Data.Params[LedcChannel].MaxWidth * _Freq) < 80000000) break;
      ResBits -= 1;
      Data.Params[LedcChannel].MaxWidth /= 2;
    }
  
    //  esp_err_t ledc_set_freq(ledc_mode_t speed_mode, ledc_timer_t timer_num, uint32_t freq_hz)
    
    //  prepare and then apply the ledc timer configuration
    ledc_timer.speed_mode      = LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    ledc_timer.timer_num       = (ledc_timer_t)((LedcChannel / 2) % 4);
    ledc_timer.duty_resolution = (ledc_timer_bit_t)ResBits;
    ledc_timer.freq_hz         = _Freq;
    ESP_ERROR_CHECK(ledc_timer_config(&ledc_timer));

    ESP_ERROR_CHECK(ledc_timer_resume(LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE, (ledc_timer_t)((LedcChannel / 2) % 4)));

    //  prepare and then apply the ledc channel configuration, it sets implicitly gpio_num GPIO as output we use for PWM signal
    ledc_channel.speed_mode = LedcChannel / 8 ? LEDC_LOW_SPEED_MODE : LEDC_HIGH_SPEED_MODE;
    //  odd channels not used
    ledc_channel.channel    = (ledc_channel_t)(LedcChannel % 8);
    ledc_channel.timer_sel  = (ledc_timer_t)((LedcChannel / 2) % 4);
    ledc_channel.intr_type  = LEDC_INTR_DISABLE;
    ledc_channel.gpio_num   = Data.Params[LedcChannel].Name;
    ledc_channel.duty       = 0;
    //  **************** set hpoint to 1 to get step GPIO low always after config *******************
    //  ********** if hpoint set to 0 step GPIO goes high after first use of ledc channel ***********
    ledc_channel.hpoint     = 1;
    ESP_ERROR_CHECK(ledc_channel_config(&ledc_channel));
  }
  return true;
}

//================================================================================================
//                                  cIotModuleCounter body
//================================================================================================

void cIotModuleCounter::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  
  //  pointer to App object used to address Generic* functions
  this->pIotModuleApp = pIotModuleApp;
    
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    Data.Params[Channel].Attached = false;
  }
}
//  called from cIotModuleApp
void cIotModuleCounter::SetControlPin(const uint8_t Channel, const gpio_num_t Name, gpio_pull_mode_t PullMode, const gpio_pull_mode_t ResetMode)
{
  Data.Params[Channel % MAX_CHANNELS].Name = Name;
  Data.Params[Channel % MAX_CHANNELS].PullMode = PullMode;
  Data.Params[Channel % MAX_CHANNELS].ResetMode = ResetMode;
}
//  called from cIotModuleApp
bool cIotModuleCounter::Attach(const uint8_t Channel)
{
  if (!Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(gpio_set_direction(Data.Params[Channel].Name, GPIO_MODE_INPUT));
    ESP_ERROR_CHECK(gpio_set_pull_mode(Data.Params[Channel].Name, Data.Params[Channel].PullMode));

    pcnt_unit.pulse_gpio_num = Data.Params[Channel].Name;
    pcnt_unit.ctrl_gpio_num = PCNT_PIN_NOT_USED;
    pcnt_unit.unit = (pcnt_unit_t)Channel;  //  unit
    pcnt_unit.channel = PCNT_CHANNEL_0;     //  channel
    pcnt_unit.pos_mode = PCNT_COUNT_INC;    //  count Only On Rising-Edges
    pcnt_unit.neg_mode = PCNT_COUNT_DIS;    //  discard Falling-Edge
    pcnt_unit.lctrl_mode = PCNT_MODE_KEEP;
    pcnt_unit.hctrl_mode = PCNT_MODE_KEEP;
    pcnt_unit.counter_h_lim = 0;
    pcnt_unit.counter_l_lim = 0;

    ESP_ERROR_CHECK(pcnt_unit_config(&pcnt_unit));
    //  pcnt_set_filter_value((pcnt_unit_t)Channel, 250);
    //  pcnt_filter_enable((pcnt_unit_t)Channel);
    pcnt_counter_pause((pcnt_unit_t)Channel);   //  pause
    pcnt_counter_clear((pcnt_unit_t)Channel);   //  clear
    pcnt_counter_resume((pcnt_unit_t)Channel);  //  resume

    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Attach - Channel : %d", __FILENAME__, __LINE__, Channel);
    Data.Params[Channel].Attached = true;
  }
  return true;
}
//  called from cIotModuleApp
bool cIotModuleCounter::Detach(const uint8_t Channel)
{
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[Channel].Name));
    Data.Params[Channel].Attached = false;
  }
  return true;
}
//  fast, called from cIotModuleApp
uint16_t IRAM_ATTR cIotModuleCounter::GetCount(const uint8_t Channel)
{
  uint16_t RetVal = 0;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)Channel, (int16_t*)&RetVal));
  }
  return RetVal;
}

//================================================================================================
//                                  cIotModuleHWEncoder body
//================================================================================================

void cIotModuleHWEncoder::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);
  
  //  pointer to App object used to address Generic* functions
  this->pIotModuleApp = pIotModuleApp;
    
  //  for every channel
  for (uint8_t Channel = 0; Channel < MAX_CHANNELS; Channel++)
  {
    Data.Params[Channel].Attached = false;
  }
}
//  called from cIotModuleApp
void cIotModuleHWEncoder::SetControlPin(const uint8_t Channel, uint8_t Type, const gpio_num_t Name, gpio_pull_mode_t PullMode, const gpio_pull_mode_t ResetMode)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - SetControlPin - Type : %d, Name : %d", __FILENAME__, __LINE__, Type, Name);
  switch(Type)
  {
    case PIN_A:
      Data.Params[Channel % MAX_CHANNELS].PinA_Name = Name;
      Data.Params[Channel % MAX_CHANNELS].PinA_PullMode = PullMode;
      Data.Params[Channel % MAX_CHANNELS].PinA_ResetMode = ResetMode;
      break;
    case PIN_B:
      Data.Params[Channel % MAX_CHANNELS].PinB_Name = Name;
      Data.Params[Channel % MAX_CHANNELS].PinB_PullMode = PullMode;
      Data.Params[Channel % MAX_CHANNELS].PinB_ResetMode = ResetMode;
      break;
  }
}
//  called from cIotModuleApp
bool cIotModuleHWEncoder::Attach(const uint8_t Channel)
{
  if (!Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(gpio_set_direction(Data.Params[Channel].PinA_Name, GPIO_MODE_INPUT));
    ESP_ERROR_CHECK(gpio_set_direction(Data.Params[Channel].PinB_Name, GPIO_MODE_INPUT));
    ESP_ERROR_CHECK(gpio_set_pull_mode(Data.Params[Channel].PinA_Name, Data.Params[Channel].PinA_PullMode));
    ESP_ERROR_CHECK(gpio_set_pull_mode(Data.Params[Channel].PinB_Name, Data.Params[Channel].PinB_PullMode));

    pcnt_unit.pulse_gpio_num = Data.Params[Channel].PinA_Name; //  rotary encoder chan A
    pcnt_unit.ctrl_gpio_num = Data.Params[Channel].PinB_Name; //  rotary encoder chan B
    
    pcnt_unit.unit = (pcnt_unit_t)Channel;      //  unit
    pcnt_unit.channel = PCNT_CHANNEL_0;         //  channel
    pcnt_unit.pos_mode = PCNT_COUNT_DEC;        //  Decrement on Rising-Edges counts x 2
    pcnt_unit.neg_mode = PCNT_COUNT_INC;        //  Increment on Falling-Edge counts x 2
    pcnt_unit.lctrl_mode = PCNT_MODE_KEEP;      //  rising A on low B = CW Step
    pcnt_unit.hctrl_mode = PCNT_MODE_REVERSE;   //  rising A on high B = CCW Step
    pcnt_unit.counter_h_lim = 0;
    pcnt_unit.counter_l_lim = 0;

    ESP_ERROR_CHECK(pcnt_unit_config(&pcnt_unit));
    //  pcnt_set_filter_value((pcnt_unit_t)Channel, 250);
    //  pcnt_filter_enable((pcnt_unit_t)Channel);
    pcnt_counter_pause((pcnt_unit_t)Channel);  //  pause
    pcnt_counter_clear((pcnt_unit_t)Channel);  //  clear
    pcnt_counter_resume((pcnt_unit_t)Channel); //  resume

    ESP_LOG_LEVEL_LOCAL(ESP_LOG_DEBUG, pTAG, "%18s %4d - Attach - Channel : %d", __FILENAME__, __LINE__, Channel);
    Data.Params[Channel].Attached = true;
  }
  return true;
}
//  called from cIotModuleApp
bool cIotModuleHWEncoder::Detach(const uint8_t Channel)
{
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[Channel].PinA_Name));
    ESP_ERROR_CHECK(gpio_reset_pin(Data.Params[Channel].PinB_Name));
    Data.Params[Channel].Attached = false;
  }
  return true;
}
//  fast, called from cIotModuleApp
uint16_t IRAM_ATTR cIotModuleHWEncoder::GetCount(const uint8_t Channel)
{
  uint16_t RetVal = 0;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    ESP_ERROR_CHECK(pcnt_get_counter_value((pcnt_unit_t)Channel, (int16_t*)&RetVal));
  }
  return RetVal;
}

//================================================================================================
//                                  cIotModuleSWEncoder body
//================================================================================================

void cIotModuleSWEncoder::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_CHANNELS; ChannelNumber++)
  {
    Data.Params[ChannelNumber].Attached = false;
  }
}

void cIotModuleSWEncoder::PreFuota(void)
{
  //  Stop any interrupt activity
  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_CHANNELS; ChannelNumber++)
  {
    if (Data.Params[ChannelNumber].Attached)
    {
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)Data.Params[ChannelNumber].PinA_Name));
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)Data.Params[ChannelNumber].PinB_Name));
      Data.Params[ChannelNumber].Attached = false;
    }
  }
}

bool cIotModuleSWEncoder::Attach(const uint8_t Channel, gpio_num_t PinA_Name, gpio_num_t PinB_Name)
{
  bool ExitValue = false;
  if (!Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    //  store params
    Data.Params[Channel].PinA_Name = PinA_Name;
    Data.Params[Channel].PinB_Name = PinB_Name;
    //  init phase
    Data.Params[Channel].PinAB_StatusOld = gpio_get_level((gpio_num_t)PinB_Name);
    Data.Params[Channel].PinAB_StatusOld <<= 1;
    Data.Params[Channel].PinAB_StatusOld += gpio_get_level((gpio_num_t)PinA_Name);
    Data.Params[Channel].PinAB_StatusOld &= 0b11;
    //  reset count
    Data.Params[Channel].Count = 0;
    //  ok
    Data.Params[Channel].Attached = true;
    //  enable GPIO interrupt on change
    ESP_ERROR_CHECK(gpio_set_intr_type(PinA_Name, GPIO_INTR_ANYEDGE));
    ESP_ERROR_CHECK(gpio_set_intr_type(PinB_Name, GPIO_INTR_ANYEDGE));
    //  add ISR handler
    void (* ISR_OnChange)(void *arg);
    switch (Channel)
    {
      case 0: ISR_OnChange = ISR_OnChange_0; break;
      case 1: ISR_OnChange = ISR_OnChange_1; break;
      case 2: ISR_OnChange = ISR_OnChange_2; break;
      case 3: ISR_OnChange = ISR_OnChange_3; break;
      case 4: ISR_OnChange = ISR_OnChange_4; break;
      case 5: ISR_OnChange = ISR_OnChange_5; break;
      case 6: ISR_OnChange = ISR_OnChange_6; break;
      case 7: ISR_OnChange = ISR_OnChange_7; break;
    }
    //  add ISR Handler
    ESP_ERROR_CHECK(gpio_isr_handler_add(PinA_Name, ISR_OnChange, this));
    ESP_ERROR_CHECK(gpio_isr_handler_add(PinB_Name, ISR_OnChange, this));
    //  enable GPIO interrupt on change
    //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinA_Name));
    //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinB_Name));
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d - PinA_Name : %d - PinB_Name : %d", __FILENAME__, __LINE__, ExitValue, Channel, PinA_Name, PinB_Name);
  }
  return ExitValue;
}

bool cIotModuleSWEncoder::Detach(const uint8_t Channel)
{
  bool ExitValue = false;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    //  disable GPIO interrupt on change
    //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)Data.Params[Channel].PinA_Name));
    //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)Data.Params[Channel].PinB_Name));
    //  remove ISR Handler
    ESP_ERROR_CHECK(gpio_isr_handler_remove(Data.Params[Channel].PinA_Name));
    ESP_ERROR_CHECK(gpio_isr_handler_remove(Data.Params[Channel].PinB_Name));

    Data.Params[Channel].Attached = false;
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}
//  fast
bool IRAM_ATTR cIotModuleSWEncoder::GetCount(const uint8_t Channel, int16_t *Count)
{
  bool ExitValue = false;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    *Count = Data.Params[Channel].Count;
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

void IRAM_ATTR cIotModuleSWEncoder::OnChangeCommon(void *arg, const uint8_t Channel)
{
  cIotModuleSWEncoder *pIotModuleSWEncoder = (cIotModuleSWEncoder *)arg;
  //  bit 3 = B(old), bit 2 = A(old), bit 1 = B(new), bit 0 = A(new)
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_StatusOld <<= 2;
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status = gpio_get_level(pIotModuleSWEncoder->Data.Params[Channel].PinB_Name);
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status <<= 1;
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status += gpio_get_level(pIotModuleSWEncoder->Data.Params[Channel].PinA_Name);
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status |= pIotModuleSWEncoder->Data.Params[Channel].PinAB_StatusOld;
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status &= 0b1111;
  //  read increment/decrememt/noaction from lookup table
  pIotModuleSWEncoder->Data.Params[Channel].Count += pIotModuleSWEncoder->QuadratureTable[pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status];
  //  store new data
  pIotModuleSWEncoder->Data.Params[Channel].PinAB_StatusOld = pIotModuleSWEncoder->Data.Params[Channel].PinAB_Status;
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_0(void *arg)
{
  //  channel 0
  OnChangeCommon(arg, 0);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_1(void *arg)
{
  //  channel 1
  OnChangeCommon(arg, 1);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_2(void *arg)
{
  //  channel 2
  OnChangeCommon(arg, 2);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_3(void *arg)
{
  //  channel 3
  OnChangeCommon(arg, 3);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_4(void *arg)
{
  //  channel 4
  OnChangeCommon(arg, 4);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_5(void *arg)
{
  //  channel 5
  OnChangeCommon(arg, 5);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_6(void *arg)
{
  //  channel 6
  OnChangeCommon(arg, 6);
}

void IRAM_ATTR cIotModuleSWEncoder::ISR_OnChange_7(void *arg)
{
  //  channel 7
  OnChangeCommon(arg, 7);
}

//================================================================================================
//                                  cIotModulePeriod body
//================================================================================================

void cIotModulePeriod::Initialize(cIotModuleApp *pIotModuleApp)
{
  ESP_LOG_LEVEL_LOCAL(ESP_LOG_INFO, pTAG, "%18s %4d - initializing", __FILENAME__, __LINE__);

  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_CHANNELS; ChannelNumber++)
  {
    Data.Params[ChannelNumber].Attached = false;
  }
}

void cIotModulePeriod::PreFuota(void)
{
  //  Stop any interrupt activity
  for (uint8_t ChannelNumber = 0; ChannelNumber < MAX_CHANNELS; ChannelNumber++)
  {
    if (Data.Params[ChannelNumber].Attached)
    {
      ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)Data.Params[ChannelNumber].PinName));
      Data.Params[ChannelNumber].Attached = false;
    }
  }
}

bool cIotModulePeriod::Attach(const uint8_t Channel, const uint8_t PinName)
{
  bool ExitValue = false;
  if (!Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    //  store params
    Data.Params[Channel].PinName = PinName;
    //  reset count
    Data.Params[Channel].Count = 0;
    //  init timer register old
    Data.Params[Channel].TimeCounterOld = 0;
    //  ok
    Data.Params[Channel].Attached = true;
    //  enable GPIO interrupt on rise
    ESP_ERROR_CHECK(gpio_set_intr_type((gpio_num_t)PinName, GPIO_INTR_POSEDGE ));
    //  add ISR handler
    void (* ISR_OnRise)(void *arg);
    switch (Channel)
    {
      case 0: ISR_OnRise = ISR_OnRise_0; break;
      case 1: ISR_OnRise = ISR_OnRise_1; break;
      case 2: ISR_OnRise = ISR_OnRise_2; break;
      case 3: ISR_OnRise = ISR_OnRise_3; break;
      case 4: ISR_OnRise = ISR_OnRise_4; break;
      case 5: ISR_OnRise = ISR_OnRise_5; break;
      case 6: ISR_OnRise = ISR_OnRise_6; break;
      case 7: ISR_OnRise = ISR_OnRise_7; break;
    }
    //  add ISR Handler
    ESP_ERROR_CHECK(gpio_isr_handler_add((gpio_num_t)PinName, ISR_OnRise, this));
    //  enable GPIO interrupt on change
    //  ESP_ERROR_CHECK(gpio_intr_enable((gpio_num_t)PinName));
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d - PinName : %d", __FILENAME__, __LINE__, ExitValue, Channel, PinName);
  }
  return ExitValue;
}

bool cIotModulePeriod::Detach(const uint8_t Channel)
{
  bool ExitValue = false;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    //  disable GPIO interrupt on change
    //  ESP_ERROR_CHECK(gpio_intr_disable((gpio_num_t)Data.Params[Channel].PinName));
    //  remove ISR Handler
    ESP_ERROR_CHECK(gpio_isr_handler_remove((gpio_num_t)Data.Params[Channel].PinName));

    Data.Params[Channel].Attached = false;
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}
//  fast
bool IRAM_ATTR cIotModulePeriod::GetCount(const uint8_t Channel, uint32_t *Count)
{
  bool ExitValue = false;
  if (Data.Params[Channel % MAX_CHANNELS].Attached)
  {
    //  read the period count
    *Count = Data.Params[Channel].Count;
    ExitValue = true;
  }
  if (ExitValue == false)
  {
    ESP_LOG_LEVEL_LOCAL(ESP_LOG_ERROR, pTAG, "%18s %4d - ExitValue : %d - Channel : %d", __FILENAME__, __LINE__, ExitValue, Channel);
  }
  return ExitValue;
}

void IRAM_ATTR cIotModulePeriod::OnRiseCommon(void *arg, const uint8_t Channel)
{
  cIotModulePeriod *pIotModulePeriod = (cIotModulePeriod *)arg;
  //  value of timer register at interrupt
  //  *** IotHAL processing must be modified according to the frequency, now 240 MHz base frequency ***

  //  first option (1 MHz base frequency):
  //  uint32_t TimeCounter = micros() * 10;
  //  second option (XTHAL_GET_CCOUNT macro and inline code are exactly the same code), base frequency 240 Mhz:
  uint32_t TimeCounter = XTHAL_GET_CCOUNT();
  //  uint32_t TimeCounter = ({ int __ccount; __asm__ __volatile__("rsr.ccount %0" : "=a"(__ccount)); __ccount; });

  //  period count between rising edges
  pIotModulePeriod->Data.Params[Channel].Count = TimeCounter - pIotModulePeriod->Data.Params[Channel].TimeCounterOld;
  //  store latest reading as start for next calculation
  pIotModulePeriod->Data.Params[Channel].TimeCounterOld = TimeCounter;
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_0(void *arg)
{
  //  channel 0
  OnRiseCommon(arg, 0);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_1(void *arg)
{
  //  channel 1
  OnRiseCommon(arg, 1);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_2(void *arg)
{
  //  channel 2
  OnRiseCommon(arg, 2);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_3(void *arg)
{
  //  channel 3
  OnRiseCommon(arg, 3);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_4(void *arg)
{
  //  channel 4
  OnRiseCommon(arg, 4);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_5(void *arg)
{
  //  channel 5
  OnRiseCommon(arg, 5);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_6(void *arg)
{
  //  channel 6
  OnRiseCommon(arg, 6);
}

void IRAM_ATTR cIotModulePeriod::ISR_OnRise_7(void *arg)
{
  //  channel 7
  OnRiseCommon(arg, 7);
}
