#ifndef IotModule_h
#define IotModule_h

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <vector>

#include <esp32-hal-timer.h>    //  attachInterrupt(), RISING, CHANGE
#include "xtensa/core-macros.h" //  XTHAL_GET_CCOUNT macro is defined here
#include "nvs.h"                //  nvs_handle type
#include "driver/pcnt.h"        //  various counter functions
#include "esp_adc_cal.h"        //  esp_adc_cal_check_efuse function
#include "driver/dac.h"         //  dac functions
#include "esp_wifi.h"           //  wifi functions 
#include "esp_event_loop.h"     //  esp_event_loop_init function
#include "sys/socket.h"         //  socket functions
#include "soc/efuse_reg.h"      //  chip package functions
#include "nvs_flash.h"          //  nvs_flash_init function
#include "esp_ota_ops.h"        //  esp_ota_get_running_partition function
#include "mdns.h"               //  mdns functions
#include "MD5Builder.h"         //  md5 functions
#include "rom/rtc.h"            //  rtc_get_reset_reason function
#include "driver/spi_master.h"  //  SPI for AD7124
#include "driver/ledc.h"        //  pwm for stepper
#include "soc/ledc_reg.h"       //  pwm for stepper
#include "soc/rtc.h"       //  pwm for stepper

#include "..\IotModuleSetup.h"

#define IOTMODULE_FIRMWARE_VER 191
#define IOTMODULE_THEREMINO_IOT_PROTOCOL "TH-IOT"
#define IOTMODULE_THEREMINO_IOT_PROTOCOL_VER 2

//#define IOTMODULE_DEBUG

#ifndef IOTMODULE_LOG_LEVEL
#define IOTMODULE_LOG_LEVEL ESP_LOG_NONE
#endif

#define __FILENAME__ (strrchr(__FILE__, '\\') ? strrchr(__FILE__, '\\') + 1 : __FILE__)

#ifndef IOTMODULE_GENERIC_PINS
#define IOTMODULE_GENERIC_PINS 0
#elif ~(~IOTMODULE_GENERIC_PINS + 0) == 0 && ~(~IOTMODULE_GENERIC_PINS + 1) == 1
#undef IOTMODULE_GENERIC_PINS
#define IOTMODULE_GENERIC_PINS 0
#elif IOTMODULE_GENERIC_PINS > 20
#undef IOTMODULE_GENERIC_PINS
#define IOTMODULE_GENERIC_PINS 20
#endif

#ifndef IOTMODULE_ADC24_PINS
#define IOTMODULE_ADC24_PINS 0
#elif ~(~IOTMODULE_ADC24_PINS + 0) == 0 && ~(~IOTMODULE_ADC24_PINS + 1) == 1
#undef IOTMODULE_ADC24_PINS
#define IOTMODULE_ADC24_PINS 0
#elif IOTMODULE_ADC24_PINS > 16
#undef IOTMODULE_ADC24_PINS
#define IOTMODULE_ADC24_PINS 16
#endif

#if IOTMODULE_ADC24_PINS > 0
#define IOTMODULE_ADC24_PIN_MASTER 1
#else
#define IOTMODULE_ADC24_PIN_MASTER 0
#endif

#ifndef IOTMODULE_ADC24_COMMUNICATION_SPEED
#define IOTMODULE_ADC24_COMMUNICATION_SPEED 16
#elif ~(~IOTMODULE_ADC24_COMMUNICATION_SPEED + 0) == 0 && ~(~IOTMODULE_ADC24_COMMUNICATION_SPEED + 1) == 1
#undef IOTMODULE_ADC24_COMMUNICATION_SPEED
#define IOTMODULE_ADC24_COMMUNICATION_SPEED 16
#elif IOTMODULE_ADC24_COMMUNICATION_SPEED > 16
#undef IOTMODULE_ADC24_COMMUNICATION_SPEED
#define IOTMODULE_ADC24_COMMUNICATION_SPEED 16
#endif

#ifndef IOTMODULE_PINOUT_TYPE
#define IOTMODULE_PINOUT_TYPE 1
#elif ~(~IOTMODULE_PINOUT_TYPE + 0) == 0 && ~(~IOTMODULE_PINOUT_TYPE + 1) == 1
#undef IOTMODULE_PINOUT_TYPE
#define IOTMODULE_PINOUT_TYPE 1
#elif IOTMODULE_PINOUT_TYPE > 3
#undef IOTMODULE_PINOUT_TYPE
#define IOTMODULE_PINOUT_TYPE 3
#endif

#if IOTMODULE_PINOUT_TYPE == 0
#define IOTMODULE_REAL_PINS 26
#elif IOTMODULE_PINOUT_TYPE == 1
#define IOTMODULE_REAL_PINS 28
#elif IOTMODULE_PINOUT_TYPE == 2
#define IOTMODULE_REAL_PINS 28
#elif IOTMODULE_PINOUT_TYPE == 3
#define IOTMODULE_REAL_PINS 28
#endif

#ifndef IOTMODULE_INOUT_PINS
#define IOTMODULE_INOUT_PINS IOTMODULE_REAL_PINS
#elif ~(~IOTMODULE_INOUT_PINS + 0) == 0 && ~(~IOTMODULE_INOUT_PINS + 1) == 1
#undef IOTMODULE_INOUT_PINS
#define IOTMODULE_INOUT_PINS IOTMODULE_REAL_PINS
#elif IOTMODULE_INOUT_PINS > IOTMODULE_REAL_PINS
#undef IOTMODULE_INOUT_PINS
#define IOTMODULE_INOUT_PINS IOTMODULE_REAL_PINS
#endif

#ifndef IOTMODULE_OUTPUT_WATCHDOG_ENABLE
#define IOTMODULE_OUTPUT_WATCHDOG_ENABLE 0
#elif ~(~IOTMODULE_OUTPUT_WATCHDOG_ENABLE + 0) == 0 && ~(~IOTMODULE_OUTPUT_WATCHDOG_ENABLE + 1) == 1
#undef IOTMODULE_OUTPUT_WATCHDOG_ENABLE
#define IOTMODULE_OUTPUT_WATCHDOG_ENABLE 1
#elif IOTMODULE_OUTPUT_WATCHDOG_ENABLE > 1
#undef IOTMODULE_OUTPUT_WATCHDOG_ENABLE
#define IOTMODULE_OUTPUT_WATCHDOG_ENABLE 1
#endif

#ifndef IOTMODULE_FUOTA_ENABLE
#define IOTMODULE_FUOTA_ENABLE 0
#elif ~(~IOTMODULE_FUOTA_ENABLE + 0) == 0 && ~(~IOTMODULE_FUOTA_ENABLE + 1) == 1
#undef IOTMODULE_FUOTA_ENABLE
#define IOTMODULE_FUOTA_ENABLE 1
#elif IOTMODULE_FUOTA_ENABLE > 1
#undef IOTMODULE_FUOTA_ENABLE
#define IOTMODULE_FUOTA_ENABLE 1
#endif

#ifndef IOTMODULE_PHY_MODE
#define IOTMODULE_PHY_MODE "WIFI"
#endif

#ifndef IOTMODULE_WIFI_MODE
#define IOTMODULE_WIFI_MODE "SOFTAP"
#endif

#ifndef IOTMODULE_WIFI_MODE_PIN
#define IOTMODULE_WIFI_MODE_PIN -1
#endif

#ifndef IOTMODULE_WIFI_WPS_PIN
#define IOTMODULE_WIFI_WPS_PIN -1
#endif

#ifndef IOTMODULE_WIFI_WPS_PIN_LED
#define IOTMODULE_WIFI_WPS_PIN_LED -1
#endif

#ifndef IOTMODULE_WIFI_ASSOC
#define IOTMODULE_WIFI_ASSOC "SOFTAP"
#endif

#ifndef IOTMODULE_WIFI_SSID
#define IOTMODULE_WIFI_SSID "NoName"
#endif

#ifndef IOTMODULE_WIFI_PASSWORD
#define IOTMODULE_WIFI_PASSWORD "!!NoName!!"
#endif

#ifndef IOTMODULE_IP_MODE
#define IOTMODULE_IP_MODE "DHCP"
#endif

#ifndef IOTMODULE_IP_IP
#define IOTMODULE_IP_IP ""
#endif

#ifndef IOTMODULE_IP_NETWORK
#define IOTMODULE_IP_NETWORK ""
#endif

#ifndef IOTMODULE_IP_GATEWAY
#define IOTMODULE_IP_GATEWAY ""
#endif

#ifndef IOTMODULE_IP_DNS
#define IOTMODULE_IP_DNS ""
#endif

#ifndef IOTMODULE_WIFI_TX_POWER_INDEX
#define IOTMODULE_WIFI_TX_POWER_INDEX 10
#elif ~(~IOTMODULE_WIFI_TX_POWER_INDEX + 0) == 0 && ~(~IOTMODULE_WIFI_TX_POWER_INDEX + 1) == 1
#undef IOTMODULE_WIFI_TX_POWER_INDEX
#define IOTMODULE_WIFI_TX_POWER_INDEX 10
#elif IOTMODULE_WIFI_TX_POWER_INDEX > 10
#undef IOTMODULE_WIFI_TX_POWER_INDEX
#define IOTMODULE_WIFI_TX_POWER_INDEX 10
#endif

#ifndef IOTMODULE_ENCODERS_TYPE
#define IOTMODULE_ENCODERS_TYPE 0
#elif ~(~IOTMODULE_ENCODERS_TYPE + 0) == 0 && ~(~IOTMODULE_ENCODERS_TYPE + 1) == 1
#undef IOTMODULE_ENCODERS_TYPE
#define IOTMODULE_ENCODERS_TYPE 0
#elif IOTMODULE_ENCODERS_TYPE > 1
#undef IOTMODULE_ENCODERS_TYPE
#define IOTMODULE_ENCODERS_TYPE 1
#endif

#ifndef IOTMODULE_WIFI_CHANNEL
#define IOTMODULE_WIFI_CHANNEL 1
#elif ~(~IOTMODULE_WIFI_CHANNEL + 0) == 0 && ~(~IOTMODULE_WIFI_CHANNEL + 1) == 1
#undef IOTMODULE_WIFI_CHANNEL
#define IOTMODULE_WIFI_CHANNEL 1
#elif IOTMODULE_WIFI_CHANNEL > 13
#undef IOTMODULE_WIFI_CHANNEL
#define IOTMODULE_WIFI_CHANNEL 13
#endif

#ifdef IOTMODULE_STATIC_IP
#error "IOTMODULE_STATIC_IP is deprecated!"
#endif

#ifndef IOTMODULE_UDP_PORT_NUMBER
#define IOTMODULE_UDP_PORT_NUMBER 49153
#elif ~(~IOTMODULE_UDP_PORT_NUMBER + 0) == 0 && ~(~IOTMODULE_UDP_PORT_NUMBER + 1) == 1
#undef IOTMODULE_UDP_PORT_NUMBER
#define IOTMODULE_UDP_PORT_NUMBER 49153
#elif IOTMODULE_UDP_PORT_NUMBER > 65535
#undef IOTMODULE_UDP_PORT_NUMBER
#define IOTMODULE_UDP_PORT_NUMBER 49153
#endif


#ifndef IOTMODULE_STEPPER_TIMING
#define IOTMODULE_STEPPER_TIMING 0
#elif ~(~IOTMODULE_STEPPER_TIMING + 0) == 0 && ~(~IOTMODULE_STEPPER_TIMING + 1) == 1
#undef IOTMODULE_STEPPER_TIMING
#define IOTMODULE_STEPPER_TIMING 0
#elif IOTMODULE_STEPPER_TIMING > 8
#undef IOTMODULE_STEPPER_TIMING
#define IOTMODULE_STEPPER_TIMING 8
#endif

#define IOTMODULE_TOTAL_PINS IOTMODULE_REAL_PINS + IOTMODULE_ADC24_PIN_MASTER + IOTMODULE_ADC24_PINS + IOTMODULE_GENERIC_PINS
#define IOTMODULE_LOOKUP_TABLE_SIZE FIRST_GENERIC_PIN_NAME + IOTMODULE_GENERIC_PINS

/////////////////////////////////////////////////////////////

#ifdef LOG_LOCAL_LEVEL
#undef LOG_LOCAL_LEVEL
#define LOG_LOCAL_LEVEL IOTMODULE_LOG_LEVEL
#else
#define LOG_LOCAL_LEVEL IOTMODULE_LOG_LEVEL
#endif

static const char* pTAG = "";

#include "IotModuleOth-h.h"
#include "IotModuleSte-h.h"
#include "IotModuleAdc-h.h"
#include "IotModuleApp-h.h"
#include "IotModuleCom-h.h"
#include "IotModule-h.h"
#include "IotModuleSte-cpp.h"
#include "IotModuleAdc-cpp.h"
#include "IotModuleOth-cpp.h"
#include "IotModuleApp-cpp.h"
#include "IotModuleCom-cpp.h"
#include "IotModule-cpp.h"

#endif  //  #define IotModule_h
