
#include "IotModule\IotModule.h"
#include "Utility/ThereminoFilters.h"
#include <Wire.h>
#include "MAX30102.h"

MAX30102 sensor;

// ===================================
//  IotModule object, do not remove
cIotModule IotModule;
// ===================================

void setup()
{
  Serial.begin(115200);
  
  // ===================================
  //  IotModule object initialization, do not remove
  IotModule.initialize();
  // ===================================

  // ----------------------------------------------------------------
  //  This part is required for MAX3010x sensor
  // ----------------------------------------------------------------
  //  We use the IR LED only because it provides the more stable data
  //  but we light also the RED LED to see some light
  // ----------------------------------------------------------------
  Wire.begin(27, 14);                 // Initialize I2C communication (SDA, SCL, Speed)
  sensor.begin();                     // Initialize the sensor
  sensor.setLEDs(60, 60, 0);          // RED, IR and GRN LEDs - Using IR only
  sensor.setPulseWidth(pw118);        // PulseWidth = 118 uS (adc 16 bit)
  sensor.setSampleRate(sr1000);       // SampleRate = 1000 samples per second
  sensor.setWorkingMode(wmHeartRate); // wmHeartRate (red) / wmSpO2 (red and IR) / wmMultiLed (RED, IR, GRN)
  sensor.setAdcRange(rge16384);       // Adc range = max (16 uA)
  //sensor.setSamplingAverage(avg1);  // average (1 to 32) unused
}


// ------------------------------------------------- Filters declarations - HiPass 0.7 Hz
Filter hipass1(0.7, true);
Filter hipass2(0.7, true);
Filter hipass3(0.7, true);
Filter hipass4(0.7, true);
// ------------------------------------------------- Filters declarations - LoPass 3 Hz
Filter lopass1(3, false);
Filter lopass2(3, false);
Filter lopass3(3, false);
Filter lopass4(3, false);
// ------------------------------------------------- Filters declarations - LoPass 2 Hz for AutoGain
Filter lopass5(2, false);

void loop()
{
    sensor.readSensor();
    float filtered = sensor.IR;
    
    // -------------------------------------------- Hi Pass - 4 stages
    filtered = hipass1.run(filtered);
    filtered = hipass2.run(filtered);
    filtered = hipass3.run(filtered);
    filtered = hipass4.run(filtered);
    // -------------------------------------------- Low Pass - 4 stages
    filtered = lopass1.run(filtered);
    filtered = lopass2.run(filtered);
    filtered = lopass3.run(filtered);  
    filtered = lopass4.run(filtered);   
    
    // -------------------------------------------- Auto gain
    float v = abs(filtered);
    v = lopass5.run(v);
    float gain = 24000 / v;
    if (gain > 5000) gain = 5000;
    filtered *= gain;
    
    // -------------------------------------------- Limit amplitude to 16 bits
    filtered += 32768;
    if (filtered > 65535) filtered = 65535;
    if (filtered < 0) filtered = 0;
	
    // -------------------------------------------- Send raw IR and filtered IR to ArduHAL
    IotModule.genericWrite16(36, sensor.IR);
    IotModule.genericWrite16(39, filtered);


    //  TEST, EXAMPLES and DEBUG
    // -------------------------------------------- test data using Theremino-ArduHAL
    //   sensor.IR  = 66666;
    //   sensor.RED = 66667;
    //   sensor.RED = sensor.getPartID();

    // -------------------------------------------- test data using Theremino-Terminal
    //   Serial.print(String(sensor.getPartID()));
    //   Serial.write(13); 
    //   Serial.write(10);
    
    // ----------------------------------
    //  Generic Read and Write examples
    // ----------------------------------   
    //  uint32_t n;
    //  n = Theremino.genericRead8(A2);
    //  Theremino.genericWrite8(A3, n);
    //
    //  n = Theremino.genericRead16(A4);
    //  Theremino.genericWrite16(A5, n);
    //
    //  n = Theremino.genericRead24(A6);
    //  Theremino.genericWrite24(A7, n);
    
    // ----------------------------------------------------------------
    //  Delay example 
    //  a single 10 mS delay decreases drammatically the exchange rate
    // ----------------------------------------------------------------
    // delay(10);
	
    // ----------------------------------------------------------------
    //  Debug the AutoGain sending the gain to the ArduHAL using Pin 9
    // ----------------------------------------------------------------
    // Theremino.genericWrite16(9, gain);

    // ----------------------------------------------------------------
    //  Measure the loop time and send it to the ArduHAL using Pin 10
    // ----------------------------------------------------------------
    // static uint32_t lastUS;
    // uint32_t time = micros();
    // uint32_t elapsedUS = time - lastUS;
    // lastUS = time;
    // Theremino.genericWrite16(10, elapsedUS);

    // ----------------------------------------------------------------
    //  Test times
    // ----------------------------------------------------------------
    // Empty loop = 20 uS
    // 2 x Theremino.genericWrite16 = 4 uS       Tot = 24
    // sensor.readSensor = 940 uS                Tot = 964
    // amplify (1 float multiplication) = 16 uS  Tot = 980
    // limit amplitude (2 IF) = 8 uS             Tot = 988
    // 8 filters (8 x 15 float mult.) = 2000 uS  Tot = 2988

}
